from typing import Any, Iterable, Iterator, Optional, no_type_check

import polars as pl
import pyarrow as pa
from triad import Schema, make_empty_aware
from triad.utils.pyarrow import get_alter_func

from fugue import (
    ArrowDataFrame,
    DataFrame,
    IterableArrowDataFrame,
    LocalDataFrameIterableDataFrame,
)
from fugue.dev import LocalDataFrameParam, fugue_annotated_param
from .polars_dataframe import PolarsDataFrame
from fugue.plugins import as_fugue_dataset


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, pl.DataFrame))
def _pl_as_fugue_df(df: pl.DataFrame, **kwargs: Any) -> PolarsDataFrame:
    return PolarsDataFrame(df, **kwargs)


@fugue_annotated_param(pl.DataFrame)
class _PolarsParam(LocalDataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        return pl.from_arrow(df.as_arrow())

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert isinstance(output, pl.DataFrame)
        return _to_adf(output, schema=schema)

    def count(self, df: Any) -> int:  # pragma: no cover
        return df.shape[0]

    def format_hint(self) -> Optional[str]:
        return "pyarrow"


@fugue_annotated_param(
    Iterable[pl.DataFrame],
    matcher=lambda x: x == Iterable[pl.DataFrame] or x == Iterator[pl.DataFrame],
)
class _IterablePolarsParam(LocalDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> Iterable[pa.Table]:
        if not isinstance(df, LocalDataFrameIterableDataFrame):
            yield pl.from_arrow(df.as_arrow())
        else:  # pragma: no cover # spark code coverage can't be included
            for sub in df.native:
                yield pl.from_arrow(sub.as_arrow())

    @no_type_check
    def to_output_df(
        self, output: Iterable[pl.DataFrame], schema: Any, ctx: Any
    ) -> DataFrame:
        def dfs(_schema: Schema) -> Iterable[ArrowDataFrame]:
            if output is not None:
                for df in output:
                    yield _to_adf(df, _schema)

        _schema: Optional[Schema] = (
            None
            if schema is None
            else (schema if isinstance(schema, Schema) else Schema(schema))
        )
        _dfs = make_empty_aware(dfs(_schema))
        if not _dfs.empty:
            return IterableArrowDataFrame(_dfs)
        return IterableArrowDataFrame([], schema=_schema)

    @no_type_check
    def count(self, df: Iterable[pl.DataFrame]) -> int:  # pragma: no cover
        return sum(_.shape[0] for _ in df)

    def format_hint(self) -> Optional[str]:
        return "pyarrow"


def _to_adf(output: pl.DataFrame, schema: Any) -> ArrowDataFrame:
    adf = output.to_arrow()
    if schema is None:  # pragma: no cover
        return ArrowDataFrame(adf)
    _schema = schema if isinstance(schema, Schema) else Schema(schema)
    f = get_alter_func(adf.schema, _schema.pa_schema, safe=False)
    return ArrowDataFrame(f(adf))



from typing import Any, Dict, List, Optional

from triad.utils.assertion import assert_or_throw

from ..collections.yielded import Yielded
from ..constants import FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT
from ..dataframe import DataFrame, AnyDataFrame
from ..dataframe.api import get_native_as_df
from ..exceptions import FugueInterfacelessError, FugueWorkflowCompileError
from ..execution import make_execution_engine
from .workflow import FugueWorkflow


def _check_valid_input(df: Any, save_path: Optional[str]) -> None:
    # Check valid input
    if isinstance(df, str):
        assert_or_throw(
            (".csv" not in df) and (".json" not in df),
            FugueInterfacelessError(
                """Fugue transform can only load parquet file paths.
                Csv and json are disallowed"""
            ),
        )
    if save_path:
        assert_or_throw(
            (".csv" not in save_path) and (".json" not in save_path),
            FugueInterfacelessError(
                """Fugue transform can only load parquet file paths.
                Csv and json are disallowed"""
            ),
        )


def transform(  # noqa: C901
    df: Any,
    using: Any,
    schema: Any = None,
    params: Any = None,
    partition: Any = None,
    callback: Any = None,
    ignore_errors: Optional[List[Any]] = None,
    persist: bool = False,
    as_local: bool = False,
    save_path: Optional[str] = None,
    checkpoint: bool = False,
    engine: Any = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
) -> Any:
    """Transform this dataframe using transformer. It's a wrapper of
    :meth:`~fugue.workflow.workflow.FugueWorkflow.transform` and
    :meth:`~fugue.workflow.workflow.FugueWorkflow.run`. It will let you do
    the basic dataframe transformation without using
    :class:`~fugue.workflow.workflow.FugueWorkflow` and
    :class:`~fugue.dataframe.dataframe.DataFrame`. Also, only native
    types are accepted for both input and output.

    Please read |TransformerTutorial|

    :param df: |DataFrameLikeObject| or :class:`~fugue.workflow.yielded.Yielded`
        or a path string to a parquet file
    :param using: transformer-like object, can't be a string expression
    :param schema: |SchemaLikeObject|, defaults to None. The transformer
        will be able to access this value from
        :meth:`~fugue.extensions.context.ExtensionContext.output_schema`
    :param params: |ParamsLikeObject| to run the processor, defaults to None.
        The transformer will be able to access this value from
        :meth:`~fugue.extensions.context.ExtensionContext.params`
    :param partition: |PartitionLikeObject|, defaults to None
    :param callback: |RPCHandlerLikeObject|, defaults to None
    :param ignore_errors: list of exception types the transformer can ignore,
        defaults to None (empty list)
    :param engine: it can be empty string or null (use the default execution
        engine), a string (use the registered execution engine), an
        :class:`~fugue.execution.execution_engine.ExecutionEngine` type, or
        the :class:`~fugue.execution.execution_engine.ExecutionEngine` instance
        , or a tuple of two values where the first value represents execution
        engine and the second value represents the sql engine (you can use ``None``
        for either of them to use the default one), defaults to None
    :param engine_conf: |ParamsLikeObject|, defaults to None
    :param as_fugue: If true, the function will always return
        a ``FugueDataFrame``, otherwise, if ``df`` is in native dataframe types such
        as pandas dataframe, then the output will also return in its native format.
        Defaults to False
    :param persist: Whether to persist(materialize) the dataframe before returning
    :param as_local: If true, the result will be converted to a ``LocalDataFrame``
    :param save_path: Whether to save the output to a file (see the note)
    :param checkpoint: Whether to add a checkpoint for the output (see the note)

    :return: the transformed dataframe, if ``df`` is a native dataframe (e.g.
        pd.DataFrame, spark dataframe, etc), the output will be a native dataframe,
        the type is determined by the execution engine you use. But if ``df`` is
        of type :class:`~fugue.dataframe.dataframe.DataFrame`, then the output will
        also be a :class:`~fugue.dataframe.dataframe.DataFrame`

    .. note::

        This function may be lazy and return the transformed dataframe.

    .. note::

        When you use callback in this function, you must be careful that the output
        dataframe must be materialized. Otherwise, if the real compute happens out of
        the function call, the callback receiver is already shut down. To do that you
        can either use ``persist`` or ``as_local``, both will materialize the dataframe
        before the callback receiver shuts down.

    .. note::

        * When `save_path` is None and `checkpoint` is False, then the output will
          not be saved into a file. The return will be a dataframe.
        * When `save_path` is None and `checkpoint` is True, then the output is
          saved into the path set by `fugue.workflow.checkpoint.path`, the name will
          be randomly chosen, and it is NOT a deterministic checkpoint, so if you run
          multiple times, the output will be saved into different files. The return
          will be a dataframe.
        * When `save_path` is not None and `checkpoint` is False, then the output will
          be saved into `save_path`. The return will be the value of `save_path`
        * When `save_path` is not None and `checkpoint` is True, then the output will
          be saved into `save_path`. The return will be the dataframe from `save_path`

        This function can only take parquet file paths in `df` and `save_path`.
        Csv and other file formats are disallowed.

        The checkpoint here is NOT deterministic, so re-run will generate new
        checkpoints.

        If you want to read and write other file formats or if you want to use
        deterministic checkpoints, please use
        :class:`~fugue.workflow.workflow.FugueWorkflow`.
    """
    _check_valid_input(df, save_path)

    dag = FugueWorkflow(compile_conf={FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT: 0})
    try:
        src = dag.create(df)
    except FugueWorkflowCompileError:
        if isinstance(df, str):
            src = dag.load(df, fmt="parquet")
        else:
            raise
    tdf = src.transform(
        using=using,
        schema=schema,
        params=params,
        pre_partition=partition,
        callback=callback,
        ignore_errors=ignore_errors or [],
    )
    if persist:
        tdf = tdf.persist()
    if checkpoint:
        if save_path is None:

            def _no_op_processor(df: DataFrame) -> DataFrame:
                # this is a trick to force yielding again
                # from the file to a dataframe
                return df

            tdf.yield_file_as("file_result")
            tdf.process(_no_op_processor).yield_dataframe_as(
                "result", as_local=as_local
            )
        else:
            tdf.save_and_use(save_path, fmt="parquet").yield_dataframe_as(
                "result", as_local=as_local
            )
    else:
        if save_path is None:
            tdf.yield_dataframe_as("result", as_local=as_local)
        else:
            tdf.save(save_path, fmt="parquet")

    dag.run(make_execution_engine(engine, conf=engine_conf, infer_by=[df]))
    if checkpoint:
        result = dag.yields["result"].result  # type:ignore
    else:
        if save_path is None:
            result = dag.yields["result"].result  # type:ignore
        else:
            return save_path
    if as_fugue or isinstance(df, (DataFrame, Yielded)):
        return result
    return result.as_pandas() if result.is_local else result.native  # type:ignore


def out_transform(
    df: Any,
    using: Any,
    params: Any = None,
    partition: Any = None,
    callback: Any = None,
    ignore_errors: Optional[List[Any]] = None,
    engine: Any = None,
    engine_conf: Any = None,
) -> None:
    """Transform this dataframe using transformer. It's a wrapper of
    :meth:`~fugue.workflow.workflow.FugueWorkflow.out_transform` and
    :meth:`~fugue.workflow.workflow.FugueWorkflow.run`. It will let you do the
    basic dataframe transformation without using
    :class:`~fugue.workflow.workflow.FugueWorkflow` and
    :class:`~fugue.dataframe.dataframe.DataFrame`. Only native types are
    accepted for both input and output.

    Please read |TransformerTutorial|

    :param df: |DataFrameLikeObject| or :class:`~fugue.workflow.yielded.Yielded`
        or a path string to a parquet file
    :param using: transformer-like object, can't be a string expression
    :param params: |ParamsLikeObject| to run the processor, defaults to None
        The transformer will be able to access this value from
        :meth:`~fugue.extensions.context.ExtensionContext.params`
    :param partition: |PartitionLikeObject|, defaults to None
    :param callback: |RPCHandlerLikeObject|, defaults to None
    :param ignore_errors: list of exception types the transformer can ignore,
        defaults to None (empty list)
    :param engine: it can be empty string or null (use the default execution
        engine), a string (use the registered execution engine), an
        :class:`~fugue.execution.execution_engine.ExecutionEngine` type, or
        the :class:`~fugue.execution.execution_engine.ExecutionEngine` instance
        , or a tuple of two values where the first value represents execution
        engine and the second value represents the sql engine (you can use ``None``
        for either of them to use the default one), defaults to None
    :param engine_conf: |ParamsLikeObject|, defaults to None

    .. note::

        This function can only take parquet file paths in `df`. CSV and JSON file
        formats are disallowed.

        This transformation is guaranteed to execute immediately (eager)
        and return nothing
    """
    dag = FugueWorkflow(compile_conf={FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT: 0})
    try:
        src = dag.create(df)
    except FugueWorkflowCompileError:
        if isinstance(df, str):
            src = dag.load(df, fmt="parquet")
        else:
            raise
    src.out_transform(
        using=using,
        params=params,
        pre_partition=partition,
        callback=callback,
        ignore_errors=ignore_errors or [],
    )

    dag.run(make_execution_engine(engine, conf=engine_conf, infer_by=[df]))


def raw_sql(
    *statements: Any,
    engine: Any = None,
    engine_conf: Any = None,
    as_fugue: bool = False,
    as_local: bool = False,
) -> AnyDataFrame:
    """Run raw SQL on the execution engine

    :param statements: a sequence of sub-statements in string
        or dataframe-like objects
    :param engine: an engine like object, defaults to None
    :param engine_conf: the configs for the engine, defaults to None
    :param as_fugue: whether to force return a Fugue DataFrame, defaults to False
    :param as_local: whether return a local dataframe, defaults to False

    :return: the result dataframe

    .. caution::

        Currently, only ``SELECT`` statements are supported

    .. admonition:: Examples

        .. code-block:: python

            import pandas as pd
            import fugue.api as fa

            with fa.engine_context("duckdb"):
                a = fa.as_fugue_df([[0,1]], schema="a:long,b:long")
                b = pd.DataFrame([[0,10]], columns=["a","b"])
                c = fa.raw_sql("SELECT * FROM",a,"UNION SELECT * FROM",b)
                fa.as_pandas(c)
    """
    dag = FugueWorkflow(compile_conf={FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT: 0})
    sp: List[Any] = []
    infer_by: List[Any] = []
    inputs: Dict[int, Any] = {}
    for x in statements:
        if isinstance(x, str):
            sp.append(x)
        else:
            if id(x) in inputs:
                sp.append(inputs[id(x)])
            else:
                inputs[id(x)] = dag.create(x)
                sp.append(inputs[id(x)])
                infer_by.append(x)

    engine = make_execution_engine(engine, engine_conf, infer_by=infer_by)
    dag.select(*sp).yield_dataframe_as("result", as_local=as_local)
    res = dag.run(engine)

    return res["result"] if as_fugue else get_native_as_df(res["result"])



from typing import Any, Dict, Iterable, List, Optional

import pandas as pd
import pyarrow as pa
import pyspark.sql as ps
from pyspark.sql.functions import col
from triad import SerializableRLock
from triad.collections.schema import SchemaError
from triad.utils.assertion import assert_or_throw

from fugue.dataframe import (
    ArrowDataFrame,
    DataFrame,
    IterableDataFrame,
    LocalBoundedDataFrame,
)
from fugue.dataframe.utils import pa_table_as_array, pa_table_as_dicts
from fugue.exceptions import FugueDataFrameOperationError
from fugue.plugins import (
    as_array,
    as_array_iterable,
    as_arrow,
    as_dict_iterable,
    as_dicts,
    as_local_bounded,
    as_pandas,
    count,
    drop_columns,
    get_column_names,
    get_num_partitions,
    head,
    is_bounded,
    is_df,
    is_empty,
    is_local,
    rename,
    select_columns,
)

from ._utils.convert import (
    to_arrow,
    to_cast_expression,
    to_pandas,
    to_schema,
    to_type_safe_input,
)
from ._utils.misc import is_spark_connect, is_spark_dataframe


class SparkDataFrame(DataFrame):
    """DataFrame that wraps Spark DataFrame. Please also read
    |DataFrameTutorial| to understand this Fugue concept

    :param df: :class:`spark:pyspark.sql.DataFrame`
    :param schema: |SchemaLikeObject| or :class:`spark:pyspark.sql.types.StructType`,
      defaults to None.

    .. note::

        * You should use :meth:`fugue_spark.execution_engine.SparkExecutionEngine.to_df`
          instead of construction it by yourself.
        * If ``schema`` is set, then there will be type cast on the Spark DataFrame if
          the schema is different.
    """

    def __init__(self, df: Any = None, schema: Any = None):  # noqa: C901
        self._lock = SerializableRLock()
        if is_spark_dataframe(df):
            if schema is not None:
                schema = to_schema(schema).assert_not_empty()
                has_cast, expr = to_cast_expression(df, schema, True)
                if has_cast:
                    df = df.selectExpr(*expr)  # type: ignore
            else:
                schema = to_schema(df).assert_not_empty()
            self._native = df
            super().__init__(schema)
        else:  # pragma: no cover
            assert_or_throw(schema is not None, SchemaError("schema is None"))
            schema = to_schema(schema).assert_not_empty()
            raise ValueError(f"{df} is incompatible with SparkDataFrame")

    @property
    def alias(self) -> str:
        return "_" + str(id(self.native))

    @property
    def native(self) -> ps.DataFrame:
        """The wrapped Spark DataFrame

        :rtype: :class:`spark:pyspark.sql.DataFrame`
        """
        return self._native

    def native_as_df(self) -> ps.DataFrame:
        return self._native

    @property
    def is_local(self) -> bool:
        return False

    @property
    def is_bounded(self) -> bool:
        return True

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        res = ArrowDataFrame(self.as_arrow())
        if self.has_metadata:
            res.reset_metadata(self.metadata)
        return res

    @property
    def num_partitions(self) -> int:
        return _spark_num_partitions(self.native)

    @property
    def empty(self) -> bool:
        return self._first is None

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        return self._first  # type: ignore

    def count(self) -> int:
        with self._lock:
            if "_df_count" not in self.__dict__:
                self._df_count = self.native.count()
            return self._df_count

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        cols = (self.schema - cols).names
        return self._select_cols(cols)

    def _select_cols(self, cols: List[Any]) -> DataFrame:
        schema = self.schema.extract(cols)
        return SparkDataFrame(self.native[schema.names])

    def as_pandas(self) -> pd.DataFrame:
        return _spark_df_as_pandas(self.native)

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        return _spark_df_as_arrow(self.native)

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            self.schema.rename(columns)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        return SparkDataFrame(_rename_spark_dataframe(self.native, columns))

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self.schema.alter(columns)
        if new_schema == self.schema:
            return self
        return SparkDataFrame(self.native, new_schema)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return _spark_as_array(self.native, columns=columns, type_safe=type_safe)

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        yield from _spark_as_array_iterable(
            self.native, columns=columns, type_safe=type_safe
        )

    def as_dicts(self, columns: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        return _spark_as_dicts(self.native, columns=columns)

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None
    ) -> Iterable[Dict[str, Any]]:
        yield from _spark_as_dict_iterable(self.native, columns=columns)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        sdf = self._select_columns(columns)
        return SparkDataFrame(
            sdf.native.limit(n), sdf.schema
        ).as_local()  # type: ignore

    @property
    def _first(self) -> Optional[List[Any]]:
        with self._lock:
            if "_first_row" not in self.__dict__:
                self._first_row = self.native.first()
                if self._first_row is not None:
                    self._first_row = list(self._first_row)  # type: ignore
            return self._first_row  # type: ignore

    def _select_columns(self, columns: Optional[List[str]]) -> "SparkDataFrame":
        if columns is None:
            return self
        return SparkDataFrame(self.native.select(*columns))


@is_df.candidate(lambda df: is_spark_dataframe(df))
def _spark_is_df(df: ps.DataFrame) -> bool:
    return True


@as_arrow.candidate(lambda df: isinstance(df, ps.DataFrame))
def _spark_df_as_arrow(df: ps.DataFrame) -> pd.DataFrame:
    return to_arrow(df)


@as_pandas.candidate(lambda df: isinstance(df, ps.DataFrame))
def _spark_df_as_pandas(df: ps.DataFrame) -> pd.DataFrame:
    return to_pandas(df)


@get_num_partitions.candidate(lambda df: is_spark_dataframe(df))
def _spark_num_partitions(df: ps.DataFrame) -> int:
    return df.rdd.getNumPartitions()


@count.candidate(lambda df: is_spark_dataframe(df))
def _spark_df_count(df: ps.DataFrame) -> int:
    return df.count()


@is_bounded.candidate(lambda df: is_spark_dataframe(df))
def _spark_df_is_bounded(df: ps.DataFrame) -> bool:
    return True


@is_empty.candidate(lambda df: is_spark_dataframe(df))
def _spark_df_is_empty(df: ps.DataFrame) -> bool:
    return df.first() is None


@is_local.candidate(lambda df: is_spark_dataframe(df))
def _spark_df_is_local(df: ps.DataFrame) -> bool:
    return False


@as_local_bounded.candidate(lambda df: is_spark_dataframe(df))
def _spark_df_as_local(df: ps.DataFrame) -> pd.DataFrame:
    return to_pandas(df)


@get_column_names.candidate(lambda df: is_spark_dataframe(df))
def _get_spark_df_columns(df: ps.DataFrame) -> List[Any]:
    return df.columns


@rename.candidate(lambda df, *args, **kwargs: is_spark_dataframe(df))
def _rename_spark_df(
    df: ps.DataFrame, columns: Dict[str, Any], as_fugue: bool = False
) -> ps.DataFrame:
    if len(columns) == 0:
        return df
    _assert_no_missing(df, columns.keys())
    return _adjust_df(_rename_spark_dataframe(df, columns), as_fugue=as_fugue)


@drop_columns.candidate(lambda df, *args, **kwargs: is_spark_dataframe(df))
def _drop_spark_df_columns(
    df: ps.DataFrame, columns: List[str], as_fugue: bool = False
) -> Any:
    cols = [x for x in df.columns if x not in columns]
    if len(cols) == 0:
        raise FugueDataFrameOperationError("cannot drop all columns")
    if len(cols) + len(columns) != len(df.columns):
        _assert_no_missing(df, columns)
    return _adjust_df(df[cols], as_fugue=as_fugue)


@select_columns.candidate(lambda df, *args, **kwargs: is_spark_dataframe(df))
def _select_spark_df_columns(
    df: ps.DataFrame, columns: List[Any], as_fugue: bool = False
) -> Any:
    if len(columns) == 0:
        raise FugueDataFrameOperationError("must select at least one column")
    _assert_no_missing(df, columns)
    return _adjust_df(df[columns], as_fugue=as_fugue)


@head.candidate(lambda df, *args, **kwargs: is_spark_dataframe(df))
def _spark_df_head(
    df: ps.DataFrame,
    n: int,
    columns: Optional[List[str]] = None,
    as_fugue: bool = False,
) -> pd.DataFrame:
    if columns is not None:
        df = df[columns]
    res = df.limit(n)
    return SparkDataFrame(res).as_local() if as_fugue else to_pandas(res)


@as_array.candidate(lambda df, *args, **kwargs: is_spark_dataframe(df))
def _spark_as_array(
    df: ps.DataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Any]:
    assert_or_throw(columns is None or len(columns) > 0, ValueError("empty columns"))
    _df = df if columns is None or len(columns) == 0 else df[columns]
    return pa_table_as_array(to_arrow(_df), columns)


@as_array_iterable.candidate(lambda df, *args, **kwargs: is_spark_dataframe(df))
def _spark_as_array_iterable(
    df: ps.DataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Any]:
    if is_spark_connect(df):  # pragma: no cover
        yield from _spark_as_array(df, columns, type_safe=type_safe)
    else:
        assert_or_throw(
            columns is None or len(columns) > 0, ValueError("empty columns")
        )
        _df = df if columns is None or len(columns) == 0 else df[columns]
        if not type_safe:
            for row in to_type_safe_input(
                _df.rdd.toLocalIterator(), to_schema(_df.schema)
            ):
                yield list(row)
        else:
            tdf = IterableDataFrame(
                _spark_as_array_iterable(_df, type_safe=False), to_schema(_df.schema)
            )
            yield from tdf.as_array_iterable(type_safe=True)


@as_dicts.candidate(lambda df, *args, **kwargs: is_spark_dataframe(df))
def _spark_as_dicts(
    df: ps.DataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Dict[str, Any]]:
    assert_or_throw(columns is None or len(columns) > 0, ValueError("empty columns"))
    _df = df if columns is None or len(columns) == 0 else df[columns]
    return pa_table_as_dicts(to_arrow(_df), columns)


@as_dict_iterable.candidate(lambda df, *args, **kwargs: is_spark_dataframe(df))
def _spark_as_dict_iterable(
    df: ps.DataFrame, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Dict[str, Any]]:
    assert_or_throw(columns is None or len(columns) > 0, ValueError("empty columns"))
    _df = df if columns is None or len(columns) == 0 else df[columns]
    cols = list(_df.columns)
    for row in _spark_as_array_iterable(_df, type_safe=type_safe):
        yield dict(zip(cols, row))


def _rename_spark_dataframe(df: ps.DataFrame, names: Dict[str, Any]) -> ps.DataFrame:
    cols: List[ps.Column] = []
    for f in df.schema:
        c = col(f.name)
        if f.name in names:
            c = c.alias(names[f.name])
        cols.append(c)
    return df.select(cols)


def _assert_no_missing(df: ps.DataFrame, columns: Iterable[Any]) -> None:
    missing = set(columns) - set(df.columns)
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")


def _adjust_df(res: ps.DataFrame, as_fugue: bool):
    return res if not as_fugue else SparkDataFrame(res)



import copy
from typing import Any, Callable, Dict, List, Optional, Type, Union, no_type_check

from triad import ParamDict, Schema
from triad.utils.assertion import assert_arg_not_none, assert_or_throw
from triad.utils.convert import get_caller_global_local_vars, to_function, to_instance
from triad.utils.hash import to_uuid

from fugue._utils.interfaceless import is_class_method, parse_output_schema_from_comment
from fugue._utils.registry import fugue_plugin
from fugue.dataframe import ArrayDataFrame, DataFrame, DataFrames, LocalDataFrame
from fugue.dataframe.function_wrapper import DataFrameFunctionWrapper
from fugue.exceptions import FugueInterfacelessError
from fugue.extensions.transformer.constants import OUTPUT_TRANSFORMER_DUMMY_SCHEMA
from fugue.extensions.transformer.transformer import CoTransformer, Transformer

from .._utils import (
    load_namespace_extensions,
    parse_validation_rules_from_comment,
    to_validation_rules,
)

_TRANSFORMER_REGISTRY = ParamDict()
_OUT_TRANSFORMER_REGISTRY = ParamDict()


@fugue_plugin
def parse_transformer(obj: Any) -> Any:
    """Parse an object to another object that can be converted to a Fugue
    :class:`~fugue.extensions.transformer.transformer.Transformer`.

    .. admonition:: Examples

        .. code-block:: python

            from fugue import Transformer, parse_transformer, FugueWorkflow
            from triad import to_uuid

            class My(Transformer):
                def __init__(self, x):
                    self.x = x

                ...

                def __uuid__(self) -> str:
                    return to_uuid(super().__uuid__(), self.x)

            @parse_transformer.candidate(
                lambda x: isinstance(x, str) and x.startswith("-*"))
            def _parse(obj):
                return My(obj)

            dag = FugueWorkflow()
            dag.df([[0]], "a:int").transform("-*abc")
            # ==  dag.df([[0]], "a:int").transform(My("-*abc"))

            dag.run()
    """
    if isinstance(obj, str) and obj in _TRANSFORMER_REGISTRY:
        return _TRANSFORMER_REGISTRY[obj]
    return obj


@fugue_plugin
def parse_output_transformer(obj: Any) -> Any:
    """Parse an object to another object that can be converted to a Fugue
    :class:`~fugue.extensions.transformer.transformer.OutputTransformer`.

    .. admonition:: Examples

        .. code-block:: python

            from fugue import Transformer, parse_output_transformer, FugueWorkflow
            from triad import to_uuid

            class My(OutputTransformer):
                def __init__(self, x):
                    self.x = x

                ...

                def __uuid__(self) -> str:
                    return to_uuid(super().__uuid__(), self.x)

            @parse_output_transformer.candidate(
                lambda x: isinstance(x, str) and x.startswith("-*"))
            def _parse(obj):
                return My(obj)

            dag = FugueWorkflow()
            dag.df([[0]], "a:int").out_transform("-*abc")
            # ==  dag.df([[0]], "a:int").out_transform(My("-*abc"))

            dag.run()
    """
    if isinstance(obj, str) and obj in _OUT_TRANSFORMER_REGISTRY:
        return _OUT_TRANSFORMER_REGISTRY[obj]
    return obj


def register_transformer(
    alias: str, obj: Any, on_dup: int = ParamDict.OVERWRITE
) -> None:
    """Register transformer with an alias.

    :param alias: alias of the transformer
    :param obj: the object that can be converted to
        :class:`~fugue.extensions.transformer.transformer.Transformer` or
        :class:`~fugue.extensions.transformer.transformer.CoTransformer`
    :param on_dup: see :meth:`triad.collections.dict.ParamDict.update`
        , defaults to ``ParamDict.OVERWRITE``

    .. tip::

        Registering an extension with an alias is particularly useful for projects
        such as libraries. This is because by using alias, users don't have to
        import the specific extension, or provide the full path of the extension.
        It can make user's code less dependent and easy to understand.

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. seealso::

        Please read |TransformerTutorial|

    .. admonition:: Examples

        Here is an example how you setup your project so your users can
        benefit from this feature. Assume your project name is ``pn``

        The transformer implementation in file ``pn/pn/transformers.py``

        .. code-block:: python

            import pandas as pd

            # schema: *
            def my_transformer(df:pd.DataFrame) -> pd.DataFrame:
                return df

        Then in ``pn/pn/__init__.py``

        .. code-block:: python

            from .transformers import my_transformer
            from fugue import register_transformer

            def register_extensions():
                register_transformer("mt", my_transformer)
                # ... register more extensions

            register_extensions()

        In users code:

        .. code-block:: python

            import pn  # register_extensions will be called
            from fugue import FugueWorkflow

            dag = FugueWorkflow()
            # use my_transformer by alias
            dag.df([[0]],"a:int").transform("mt").show()
            dag.run()
    """
    _TRANSFORMER_REGISTRY.update({alias: obj}, on_dup=on_dup)


def register_output_transformer(
    alias: str, obj: Any, on_dup: int = ParamDict.OVERWRITE
) -> None:
    """Register output transformer with an alias.

    :param alias: alias of the transformer
    :param obj: the object that can be converted to
        :class:`~fugue.extensions.transformer.transformer.OutputTransformer` or
        :class:`~fugue.extensions.transformer.transformer.OutputCoTransformer`
    :param on_dup: see :meth:`triad.collections.dict.ParamDict.update`
        , defaults to ``ParamDict.OVERWRITE``

    .. tip::

        Registering an extension with an alias is particularly useful for projects
        such as libraries. This is because by using alias, users don't have to
        import the specific extension, or provide the full path of the extension.
        It can make user's code less dependent and easy to understand.

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. seealso::

        Please read |TransformerTutorial|

    .. admonition:: Examples

        Here is an example how you setup your project so your users can
        benefit from this feature. Assume your project name is ``pn``

        The transformer implementation in file ``pn/pn/transformers.py``

        .. code-block:: python

            import pandas as pd

            def my_transformer(df:pd.DataFrame) -> None:
                df.to_parquet("<unique_path>")

        Then in ``pn/pn/__init__.py``

        .. code-block:: python

            from .transformers import my_transformer
            from fugue import register_transformer

            def register_extensions():
                register_transformer("mt", my_transformer)
                # ... register more extensions

            register_extensions()

        In users code:

        .. code-block:: python

            import pn  # register_extensions will be called
            from fugue import FugueWorkflow

            dag = FugueWorkflow()
            # use my_transformer by alias
            dag.df([[0]],"a:int").out_transform("mt")
            dag.run()
    """
    _OUT_TRANSFORMER_REGISTRY.update({alias: obj}, on_dup=on_dup)


def transformer(
    schema: Any, **validation_rules: Any
) -> Callable[[Any], "_FuncAsTransformer"]:
    """Decorator for transformers

    Please read |TransformerTutorial|
    """

    def deco(func: Callable) -> "_FuncAsTransformer":
        assert_or_throw(
            not is_class_method(func),
            NotImplementedError("transformer decorator can't be used on class methods"),
        )
        return _FuncAsTransformer.from_func(
            func, schema, validation_rules=to_validation_rules(validation_rules)
        )

    return deco


def output_transformer(
    **validation_rules: Any,
) -> Callable[[Any], "_FuncAsTransformer"]:
    """Decorator for transformers

    Please read |TransformerTutorial|
    """

    def deco(func: Callable) -> "_FuncAsOutputTransformer":
        assert_or_throw(
            not is_class_method(func),
            NotImplementedError(
                "output_transformer decorator can't be used on class methods"
            ),
        )
        return _FuncAsOutputTransformer.from_func(
            func, schema=None, validation_rules=to_validation_rules(validation_rules)
        )

    return deco


def cotransformer(
    schema: Any, **validation_rules: Any
) -> Callable[[Any], "_FuncAsCoTransformer"]:
    """Decorator for cotransformers

    Please read |CoTransformerTutorial|
    """

    def deco(func: Callable) -> "_FuncAsCoTransformer":
        assert_or_throw(
            not is_class_method(func),
            NotImplementedError(
                "cotransformer decorator can't be used on class methods"
            ),
        )
        return _FuncAsCoTransformer.from_func(
            func, schema, validation_rules=to_validation_rules(validation_rules)
        )

    return deco


def output_cotransformer(
    **validation_rules: Any,
) -> Callable[[Any], "_FuncAsCoTransformer"]:
    """Decorator for cotransformers

    Please read |CoTransformerTutorial|
    """

    def deco(func: Callable) -> "_FuncAsOutputCoTransformer":
        assert_or_throw(
            not is_class_method(func),
            NotImplementedError(
                "output_cotransformer decorator can't be used on class methods"
            ),
        )
        return _FuncAsOutputCoTransformer.from_func(
            func, schema=None, validation_rules=to_validation_rules(validation_rules)
        )

    return deco


class _FuncAsTransformer(Transformer):
    def validate_on_compile(self) -> None:
        super().validate_on_compile()
        _validate_callback(self)

    def get_output_schema(self, df: DataFrame) -> Any:
        return self._parse_schema(self._output_schema_arg, df)  # type: ignore

    def get_format_hint(self) -> Optional[str]:
        return self._format_hint  # type: ignore

    @property
    def validation_rules(self) -> Dict[str, Any]:
        return self._validation_rules  # type: ignore

    @no_type_check
    def transform(self, df: LocalDataFrame) -> LocalDataFrame:
        args = [df] + _get_callback(self)
        return self._wrapper.run(
            args, self.params, ignore_unknown=False, output_schema=self.output_schema
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._wrapper(*args, **kwargs)  # type: ignore

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(self._wrapper.__uuid__(), self._output_schema_arg)

    def _parse_schema(self, obj: Any, df: DataFrame) -> Schema:
        if callable(obj):
            return obj(df, **self.params)
        if isinstance(obj, str):
            return df.schema.transform(obj)
        if isinstance(obj, List):
            return df.schema.transform(*obj)
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def from_func(
        func: Callable, schema: Any, validation_rules: Dict[str, Any]
    ) -> "_FuncAsTransformer":
        if schema is None:
            schema = parse_output_schema_from_comment(func)
        if isinstance(schema, Schema):  # to be less strict on determinism
            schema = str(schema)
        validation_rules.update(parse_validation_rules_from_comment(func))
        assert_arg_not_none(schema, "schema")
        tr = _FuncAsTransformer()
        tr._wrapper = DataFrameFunctionWrapper(  # type: ignore
            func, "^[lspq][fF]?x*z?$", "^[lspqr]$"
        )
        tr._output_schema_arg = schema  # type: ignore
        tr._validation_rules = validation_rules  # type: ignore
        tr._uses_callback = "f" in tr._wrapper.input_code.lower()  # type: ignore
        tr._requires_callback = "F" in tr._wrapper.input_code  # type: ignore
        tr._format_hint = tr._wrapper.get_format_hint()  # type: ignore
        return tr


class _FuncAsOutputTransformer(_FuncAsTransformer):
    def validate_on_compile(self) -> None:
        super().validate_on_compile()
        _validate_callback(self)

    def get_output_schema(self, df: DataFrame) -> Any:
        return OUTPUT_TRANSFORMER_DUMMY_SCHEMA

    def get_format_hint(self) -> Optional[str]:
        return self._format_hint  # type: ignore

    @no_type_check
    def transform(self, df: LocalDataFrame) -> LocalDataFrame:
        args = [df] + _get_callback(self)
        self._wrapper.run(args, self.params, ignore_unknown=False, output=False)
        return ArrayDataFrame([], OUTPUT_TRANSFORMER_DUMMY_SCHEMA)

    @staticmethod
    def from_func(
        func: Callable, schema: Any, validation_rules: Dict[str, Any]
    ) -> "_FuncAsOutputTransformer":
        assert_or_throw(schema is None, "schema must be None for output transformers")
        validation_rules.update(parse_validation_rules_from_comment(func))
        tr = _FuncAsOutputTransformer()
        tr._wrapper = DataFrameFunctionWrapper(  # type: ignore
            func, "^[lspq][fF]?x*z?$", "^[lspnqr]$"
        )
        tr._output_schema_arg = None  # type: ignore
        tr._validation_rules = validation_rules  # type: ignore
        tr._uses_callback = "f" in tr._wrapper.input_code.lower()  # type: ignore
        tr._requires_callback = "F" in tr._wrapper.input_code  # type: ignore
        tr._format_hint = tr._wrapper.get_format_hint()  # type: ignore
        return tr


class _FuncAsCoTransformer(CoTransformer):
    def validate_on_compile(self) -> None:
        super().validate_on_compile()
        _validate_callback(self)

    def get_output_schema(self, dfs: DataFrames) -> Any:
        return self._parse_schema(self._output_schema_arg, dfs)  # type: ignore

    def get_format_hint(self) -> Optional[str]:
        return self._format_hint  # type: ignore

    @property
    def validation_rules(self) -> ParamDict:
        return self._validation_rules  # type: ignore

    @no_type_check
    def transform(self, dfs: DataFrames) -> LocalDataFrame:
        cb = _get_callback(self)
        if self._dfs_input:  # function has DataFrames input
            return self._wrapper.run(  # type: ignore
                [dfs] + cb,
                self.params,
                ignore_unknown=False,
                output_schema=self.output_schema,
            )
        if not dfs.has_key:  # input does not have key
            return self._wrapper.run(  # type: ignore
                list(dfs.values()) + cb,
                self.params,
                ignore_unknown=False,
                output_schema=self.output_schema,
            )
        else:  # input DataFrames has key
            p = dict(dfs)
            p.update(self.params)
            return self._wrapper.run(  # type: ignore
                [] + cb, p, ignore_unknown=False, output_schema=self.output_schema
            )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._wrapper(*args, **kwargs)  # type: ignore

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(
            self._wrapper.__uuid__(), self._output_schema_arg, self._dfs_input
        )

    def _parse_schema(self, obj: Any, dfs: DataFrames) -> Schema:
        if callable(obj):
            return obj(dfs, **self.params)
        if isinstance(obj, str):
            return Schema(obj)
        if isinstance(obj, List):
            s = Schema()
            for x in obj:
                s += self._parse_schema(x, dfs)
            return s
        return Schema(obj)

    @staticmethod
    def from_func(
        func: Callable, schema: Any, validation_rules: Dict[str, Any]
    ) -> "_FuncAsCoTransformer":
        assert_or_throw(
            len(validation_rules) == 0,
            NotImplementedError("CoTransformer does not support validation rules"),
        )

        if schema is None:
            schema = parse_output_schema_from_comment(func)
        if isinstance(schema, Schema):  # to be less strict on determinism
            schema = str(schema)
        if isinstance(schema, str):
            assert_or_throw(
                "*" not in schema,
                FugueInterfacelessError(
                    "* can't be used on cotransformer output schema"
                ),
            )
        assert_arg_not_none(schema, "schema")
        tr = _FuncAsCoTransformer()
        tr._wrapper = DataFrameFunctionWrapper(  # type: ignore
            func, "^(c|[lspq]+)[fF]?x*z?$", "^[lspqr]$"
        )
        tr._dfs_input = tr._wrapper.input_code[0] == "c"  # type: ignore
        tr._output_schema_arg = schema  # type: ignore
        tr._validation_rules = {}  # type: ignore
        tr._uses_callback = "f" in tr._wrapper.input_code.lower()  # type: ignore
        tr._requires_callback = "F" in tr._wrapper.input_code  # type: ignore
        tr._format_hint = tr._wrapper.get_format_hint()  # type: ignore
        return tr


class _FuncAsOutputCoTransformer(_FuncAsCoTransformer):
    def validate_on_compile(self) -> None:
        super().validate_on_compile()
        _validate_callback(self)

    def get_output_schema(self, dfs: DataFrames) -> Any:
        return OUTPUT_TRANSFORMER_DUMMY_SCHEMA

    def get_format_hint(self) -> Optional[str]:
        return self._format_hint  # type: ignore

    @no_type_check
    def transform(self, dfs: DataFrames) -> LocalDataFrame:
        cb = _get_callback(self)
        if self._dfs_input:  # function has DataFrames input
            self._wrapper.run(  # type: ignore
                [dfs] + cb,
                self.params,
                ignore_unknown=False,
                output=False,
            )
        elif not dfs.has_key:  # input does not have key
            self._wrapper.run(  # type: ignore
                list(dfs.values()) + cb,
                self.params,
                ignore_unknown=False,
                output=False,
            )
        else:  # input DataFrames has key
            p = dict(dfs)
            p.update(self.params)
            self._wrapper.run(
                [] + cb, p, ignore_unknown=False, output=False  # type: ignore
            )
        return ArrayDataFrame([], OUTPUT_TRANSFORMER_DUMMY_SCHEMA)

    @staticmethod
    def from_func(
        func: Callable, schema: Any, validation_rules: Dict[str, Any]
    ) -> "_FuncAsOutputCoTransformer":
        assert_or_throw(schema is None, "schema must be None for output cotransformers")
        assert_or_throw(
            len(validation_rules) == 0,
            NotImplementedError("CoTransformer does not support validation rules"),
        )

        tr = _FuncAsOutputCoTransformer()
        tr._wrapper = DataFrameFunctionWrapper(  # type: ignore
            func, "^(c|[lspq]+)[fF]?x*z?$", "^[lspnqr]$"
        )
        tr._dfs_input = tr._wrapper.input_code[0] == "c"  # type: ignore
        tr._output_schema_arg = None  # type: ignore
        tr._validation_rules = {}  # type: ignore
        tr._uses_callback = "f" in tr._wrapper.input_code.lower()  # type: ignore
        tr._requires_callback = "F" in tr._wrapper.input_code  # type: ignore
        tr._format_hint = tr._wrapper.get_format_hint()  # type: ignore
        return tr


def _to_transformer(
    obj: Any,
    schema: Any = None,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
    validation_rules: Optional[Dict[str, Any]] = None,
) -> Union[Transformer, CoTransformer]:
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    load_namespace_extensions(obj)
    return _to_general_transformer(
        obj=parse_transformer(obj),
        schema=schema,
        global_vars=global_vars,
        local_vars=local_vars,
        validation_rules=validation_rules,
        func_transformer_type=_FuncAsTransformer,
        func_cotransformer_type=_FuncAsCoTransformer,
    )


def _to_output_transformer(
    obj: Any,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
    validation_rules: Optional[Dict[str, Any]] = None,
) -> Union[Transformer, CoTransformer]:
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    load_namespace_extensions(obj)
    return _to_general_transformer(
        obj=parse_output_transformer(obj),
        schema=None,
        global_vars=global_vars,
        local_vars=local_vars,
        validation_rules=validation_rules,
        func_transformer_type=_FuncAsOutputTransformer,
        func_cotransformer_type=_FuncAsOutputCoTransformer,
    )


def _to_general_transformer(  # noqa: C901
    obj: Any,
    schema: Any,
    global_vars: Optional[Dict[str, Any]],
    local_vars: Optional[Dict[str, Any]],
    validation_rules: Optional[Dict[str, Any]],
    func_transformer_type: Type,
    func_cotransformer_type: Type,
) -> Union[Transformer, CoTransformer]:
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    exp: Optional[Exception] = None
    if validation_rules is None:
        validation_rules = {}
    try:
        return copy.copy(
            to_instance(
                obj, Transformer, global_vars=global_vars, local_vars=local_vars
            )
        )
    except Exception as e:
        exp = e
    try:
        return copy.copy(
            to_instance(
                obj, CoTransformer, global_vars=global_vars, local_vars=local_vars
            )
        )
    except Exception as e:
        exp = e
    try:
        f = to_function(obj, global_vars=global_vars, local_vars=local_vars)
        # this is for string expression of function with decorator
        if isinstance(f, Transformer):
            return copy.copy(f)
        # this is for functions without decorator
        return func_transformer_type.from_func(
            f, schema, validation_rules=validation_rules
        )
    except Exception as e:
        exp = e
    try:
        f = to_function(obj, global_vars=global_vars, local_vars=local_vars)
        # this is for string expression of function with decorator
        if isinstance(f, CoTransformer):
            return copy.copy(f)
        # this is for functions without decorator
        return func_cotransformer_type.from_func(
            f, schema, validation_rules=validation_rules
        )
    except Exception as e:
        exp = e
    raise FugueInterfacelessError(f"{obj} is not a valid transformer", exp)


def _validate_callback(ctx: Any) -> None:
    if ctx._requires_callback:
        assert_or_throw(
            ctx.has_callback,
            FugueInterfacelessError(f"Callback is required but not provided: {ctx}"),
        )


def _get_callback(ctx: Any) -> List[Any]:
    uses_callback = ctx._uses_callback
    requires_callback = ctx._requires_callback
    if not uses_callback:
        return []
    if requires_callback:
        assert_or_throw(
            ctx.has_callback,
            FugueInterfacelessError(f"Callback is required but not provided: {ctx}"),
        )
        return [ctx.callback]
    return [ctx.callback if ctx.has_callback else None]



from logging import Logger
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union
from uuid import uuid4

from triad import to_uuid

from fugue._utils.registry import fugue_plugin
from fugue._utils.misc import import_fsql_dependency

_TEMP_TABLE_EXPR_PREFIX = "<tmpdf:"
_TEMP_TABLE_EXPR_SUFFIX = ">"


class TempTableName:
    """Generating a temporary, random and globaly unique table name"""

    def __init__(self):
        self.key = "_" + str(uuid4())[:5].upper()

    def __repr__(self) -> str:
        return _TEMP_TABLE_EXPR_PREFIX + self.key + _TEMP_TABLE_EXPR_SUFFIX


@fugue_plugin
def transpile_sql(
    raw: str, from_dialect: Optional[str], to_dialect: Optional[str]
) -> str:
    """Transpile SQL between dialects, it should work only when both
    ``from_dialect`` and ``to_dialect`` are not None

    :param raw: the raw SQL
    :param from_dialect: the dialect of the raw SQL
    :param to_dialect: the expected dialect.
    :return: the transpiled SQL
    """
    if (
        from_dialect is not None
        and to_dialect is not None
        and from_dialect != to_dialect
    ):
        sqlglot = import_fsql_dependency("sqlglot")

        return " ".join(sqlglot.transpile(raw, read=from_dialect, write=to_dialect))
    else:
        return raw


class StructuredRawSQL:
    """The Raw SQL object containing table references and dialect information.

    :param statements: In each tuple, the first value indicates whether
        the second value is a dataframe name reference (True), or just a part
        of the statement (False)
    :param dialect: the dialect of the statements, defaults to None

    .. note::

        ``dialect`` None means no transpilation will be done when constructing
        the final sql.
    """

    def __init__(
        self, statements: Iterable[Tuple[bool, str]], dialect: Optional[str] = None
    ):
        self._statements = list(statements)
        self._dialect = dialect

    @property
    def dialect(self) -> Optional[str]:
        """The dialect of this query"""
        return self._dialect

    def __uuid__(self) -> str:
        return to_uuid(self._statements, self._dialect)

    def construct(
        self,
        name_map: Union[None, Callable[[str], str], Dict[str, str]] = None,
        dialect: Optional[str] = None,
        log: Optional[Logger] = None,
    ):
        """Construct the final SQL given the ``dialect``

        :param name_map: the name map from the original statement to
            the expected names, defaults to None. It can be a function or a
            dictionary
        :param dialect: the expected dialect, defaults to None
        :param log: the logger to log information, defaults to None
        :return: the final SQL string
        """
        nm: Any = (
            (lambda x: x)
            if name_map is None
            else name_map
            if not isinstance(name_map, dict)
            else (lambda x: name_map.get(x, x))  # type: ignore
        )
        raw_sql = " ".join(nm(tp[1]) if tp[0] else tp[1] for tp in self._statements)
        if (
            self._dialect is not None
            and dialect is not None
            and self._dialect != dialect
        ):
            tsql = transpile_sql(raw_sql, self._dialect, dialect)
            if log is not None:
                log.debug(
                    "SQL transpiled from %s to %s\n\n"
                    "Original:\n\n%s\n\nTranspiled:\n\n%s\n",
                    self._dialect,
                    dialect,
                    raw_sql,
                    tsql,
                )
            return tsql
        return raw_sql

    @staticmethod
    def from_expr(
        sql: str,
        prefix: str = _TEMP_TABLE_EXPR_PREFIX,
        suffix: str = _TEMP_TABLE_EXPR_SUFFIX,
        dialect: Optional[str] = None,
    ) -> "StructuredRawSQL":
        """Parse the ``StructuredRawSQL`` from the ``sql`` expression.
        The sql should look like ``SELECT * FROM <tmpdf:dfname>``. This
        function can identify the tmpdfs with the given syntax, and construct
        the ``StructuredRawSQL``

        :param sql: the SQL expression with ``<tmpdf:?>``
        :param prefix: the prefix of the temp df
        :param suffix: the suffix of the temp df
        :param dialect: the dialect of the sql expression, defaults to None
        :return: the parsed object
        """

        def _get() -> Iterable[Tuple[bool, str]]:
            p = 0
            while p < len(sql):
                b = sql.find(prefix, p)
                if b >= 0:
                    if b > p:
                        yield (False, sql[p:b])
                    b += len(prefix)
                    e = sql.find(suffix, b)
                    yield (True, sql[b:e])
                    p = e + len(suffix)
                else:
                    yield (False, sql[p:])
                    return

        return StructuredRawSQL(_get(), dialect=dialect)



from typing import Any, Dict, Tuple

import pyarrow as pa
import pytest
from triad.utils.pyarrow import to_pa_datatype

_FUGUE_TEST_CONF_NAME = "fugue_test_conf"


def pytest_addoption(parser: Any):  # pragma: no cover
    parser.addini(
        _FUGUE_TEST_CONF_NAME,
        help="Configs for fugue testing execution engines",
        type="linelist",
    )


def pytest_configure(config: Any):
    from fugue.test.plugins import _set_global_conf

    options = config.getini(_FUGUE_TEST_CONF_NAME)
    conf: Dict[str, Any] = {}
    if options:
        for line in options:
            line = line.strip()
            if not line.startswith("#"):
                k, v = _parse_line(line)
                conf[k] = v
    _set_global_conf(conf)


def pytest_report_header(config, start_path):
    from fugue.test.plugins import _get_all_ini_conf

    header_lines = []
    header_lines.append("Fugue tests will be initialized with options:")
    for k, v in _get_all_ini_conf().items():
        header_lines.append(f"\t{k} = {v}")
    return "\n".join(header_lines)


def _parse_line(line: str) -> Tuple[str, Any]:
    try:
        kv = line.split("=", 1)
        if len(kv) == 1:
            raise ValueError()
        kt = kv[0].split(":", 1)
        if len(kt) == 1:
            tp = pa.string()
        else:
            tp = to_pa_datatype(kt[1].strip())
        key = kt[0].strip()
        if key == "":
            raise ValueError()
        value = pa.compute.cast([kv[1].strip()], tp).to_pylist()[0]
        return key, value
    except Exception:
        raise ValueError(
            f"Invalid config line: {line}, it must be in format: key[:type]=value"
        )


@pytest.fixture(scope="class")
def backend_context(request: Any):
    from fugue.test.plugins import _make_backend_context, _parse_backend

    c, _ = _parse_backend(request.param)
    session = request.getfixturevalue(c + "_session")
    with _make_backend_context(request.param, session) as ctx:
        yield ctx


@pytest.fixture(scope="class")
def _class_backend_context(request, backend_context):
    from fugue.test.plugins import FugueTestContext

    request.cls._test_context = FugueTestContext(
        engine=backend_context.engine,
        session=backend_context.session,
        name=backend_context.name,
    )
    yield



from contextlib import contextmanager
from typing import Any, Dict, Iterator

from .plugins import FugueTestBackend, fugue_test_backend


@fugue_test_backend
class PandasTestBackend(FugueTestBackend):
    name = "pandas"

    @classmethod
    @contextmanager
    def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
        yield "pandas"  # pragma: no cover


@fugue_test_backend
class NativeTestBackend(FugueTestBackend):
    name = "native"

    @classmethod
    @contextmanager
    def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
        yield "native"  # pragma: no cover



from contextlib import contextmanager
from typing import Any, Dict, Iterator

from pyspark.sql import SparkSession

import fugue.test as ft

from ._utils.misc import SparkConnectSession


@ft.fugue_test_backend
class SparkTestBackend(ft.FugueTestBackend):
    name = "spark"
    default_session_conf = {
        "spark.app.name": "fugue-test-spark",
        "spark.master": "local[*]",
        "spark.default.parallelism": 4,
        "spark.dynamicAllocation.enabled": "false",
        "spark.executor.cores": 4,
        "spark.executor.instances": 1,
        "spark.io.compression.codec": "lz4",
        "spark.rdd.compress": "false",
        "spark.sql.shuffle.partitions": 4,
        "spark.shuffle.compress": "false",
        "spark.sql.catalogImplementation": "in-memory",
        "spark.sql.execution.arrow.pyspark.enabled": True,
        "spark.sql.adaptive.enabled": False,
    }

    @classmethod
    def transform_session_conf(cls, conf: Dict[str, Any]) -> Dict[str, Any]:
        return ft.extract_conf(conf, "spark.", remove_prefix=False)

    @classmethod
    @contextmanager
    def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
        with _create_session(session_conf).getOrCreate() as spark:
            yield spark


if SparkConnectSession is not None:

    @ft.fugue_test_backend
    class SparkConnectTestBackend(SparkTestBackend):
        name = "sparkconnect"
        default_session_conf = {
            "spark.default.parallelism": 4,
            "spark.sql.shuffle.partitions": 4,
            "spark.sql.execution.arrow.pyspark.enabled": True,
            "spark.sql.adaptive.enabled": False,
        }

        @classmethod
        def transform_session_conf(
            cls, conf: Dict[str, Any]
        ) -> Dict[str, Any]:  # pragma: no cover
            # replace sparkconnect. with spark.
            return {
                "spark." + k: v
                for k, v in ft.extract_conf(
                    conf, cls.name + ".", remove_prefix=True
                ).items()
            }

        @classmethod
        @contextmanager
        def session_context(
            cls, session_conf: Dict[str, Any]
        ) -> Iterator[Any]:  # pragma: no cover
            spark = _create_session(session_conf).remote("sc://localhost").getOrCreate()
            yield spark


def _create_session(conf: Dict[str, Any]) -> Any:
    sb = SparkSession.builder
    for k, v in conf.items():
        sb = sb.config(k, v)
    return sb



from typing import Any, Dict, Iterable, List, Optional

from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.dataframe import (
    DataFrame,
    LocalBoundedDataFrame,
    LocalUnboundedDataFrame,
    _get_schema_change,
)
from fugue.exceptions import FugueDataFrameOperationError
from triad.collections.schema import Schema
from triad.utils.iter import EmptyAwareIterable, make_empty_aware
from triad.utils.pyarrow import apply_schema


class IterableDataFrame(LocalUnboundedDataFrame):
    """DataFrame that wraps native python iterable of arrays. Please read
    |DataFrameTutorial| to understand the concept

    :param df: 2-dimensional array, iterable of arrays, or
      :class:`~fugue.dataframe.dataframe.DataFrame`
    :param schema: |SchemaLikeObject|

    .. admonition:: Examples

        >>> a = IterableDataFrame([[0,'a'],[1,'b']],"a:int,b:str")
        >>> b = IterableDataFrame(a)

    .. note::

        It's ok to peek the dataframe, it will not affect the iteration, but it's
        invalid operation to count
    """

    def __init__(self, df: Any = None, schema: Any = None):  # noqa: C901
        if df is None:
            idf: Iterable[Any] = []
            orig_schema: Optional[Schema] = None
        elif isinstance(df, IterableDataFrame):
            idf = df.native
            orig_schema = df.schema
        elif isinstance(df, DataFrame):
            idf = df.as_array_iterable(type_safe=False)
            orig_schema = df.schema
        elif isinstance(df, Iterable):
            idf = df
            orig_schema = None
        else:
            raise ValueError(f"{df} is incompatible with IterableDataFrame")
        schema, pos = _get_schema_change(orig_schema, schema)
        super().__init__(schema)
        self._pos = pos
        self._native = make_empty_aware(self._preprocess(idf))

    @property
    def native(self) -> EmptyAwareIterable[Any]:
        """Iterable of native python arrays"""
        return self._native

    @property
    def empty(self) -> bool:
        return self.native.empty

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        return list(self.native.peek())

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        res: List[Any] = []
        for row in self.as_array_iterable(columns, type_safe=True):
            if n < 1:
                break
            res.append(list(row))
            n -= 1
        return ArrayDataFrame(
            res, self.schema if columns is None else self.schema.extract(columns)
        )

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        return IterableDataFrame(self, self.schema - cols)

    def _select_cols(self, keys: List[Any]) -> DataFrame:
        return IterableDataFrame(self, self.schema.extract(keys))

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            schema = self.schema.rename(columns)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        return IterableDataFrame(self.native, schema)

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self.schema.alter(columns)
        if new_schema == self.schema:
            return self
        return IterableDataFrame(self.native, new_schema)

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        return ArrayDataFrame(self.as_array(), schema=self.schema)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return list(self.as_array_iterable(columns, type_safe=type_safe))

    def as_dicts(self, columns: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        return list(self.as_dict_iterable(columns))

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        if columns is None:
            if not type_safe:
                for item in self.native:
                    yield item
            else:
                for item in apply_schema(
                    self.schema.pa_schema,
                    self.native,
                    copy=True,
                    deep=True,
                    str_as_json=True,
                ):
                    yield item
        else:
            df = IterableDataFrame(self, self.schema.extract(columns))
            for item in df.as_array_iterable(type_safe=type_safe):
                yield item

    def _preprocess(self, raw_data: Iterable[Any]) -> Iterable[Any]:
        if len(self._pos) == 0:
            for x in raw_data:
                yield x
        else:
            for x in raw_data:
                yield [x[i] for i in self._pos]



from typing import Any

try:
    from pyspark.sql.connect.session import SparkSession as SparkConnectSession
    from pyspark.sql.connect.dataframe import DataFrame as SparkConnectDataFrame
except Exception:  # pragma: no cover
    SparkConnectSession = None
    SparkConnectDataFrame = None
import pyspark.sql as ps


def is_spark_connect(session: Any) -> bool:
    return SparkConnectSession is not None and isinstance(
        session, (SparkConnectSession, SparkConnectDataFrame)
    )


def is_spark_dataframe(df: Any) -> bool:
    return isinstance(df, ps.DataFrame) or (
        SparkConnectDataFrame is not None and isinstance(df, SparkConnectDataFrame)
    )


def is_spark_session(session: Any) -> bool:
    return isinstance(session, ps.SparkSession) or (
        SparkConnectSession is not None and isinstance(session, SparkConnectSession)
    )



from contextlib import contextmanager
from typing import Any, Dict, Iterator

import dask
from dask.distributed import Client

import fugue.test as ft


@ft.fugue_test_backend
class DaskTestBackend(ft.FugueTestBackend):
    name = "dask"

    @classmethod
    def transform_session_conf(cls, conf: Dict[str, Any]) -> Dict[str, Any]:
        return ft.extract_conf(conf, "dask.", remove_prefix=True)

    @classmethod
    @contextmanager
    def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
        with Client(**session_conf) as client:
            dask.config.set({"dataframe.shuffle.method": "tasks"})
            dask.config.set({"dataframe.convert-string": False})
            yield client



from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type
from fugue.dataframe.utils import _df_eq
from triad import assert_or_throw, run_once
from triad.utils.entry_points import load_entry_point

try:
    import pytest

    _HAS_PYTEST = True
except ImportError:  # pragma: no cover
    _HAS_PYTEST = False


_FUGUE_TEST_BACKENDS: Dict[str, Type["FugueTestBackend"]] = {}
_FUGUE_TEST_ALL_INI_CONF: Dict[str, Any] = {}
_FUGUE_TEST_INI_FUGUE_CONF: Dict[str, Any] = {}


def _set_global_conf(conf: Dict[str, Any]) -> None:
    global _FUGUE_TEST_ALL_INI_CONF, _FUGUE_TEST_INI_FUGUE_CONF
    _FUGUE_TEST_ALL_INI_CONF = conf
    _FUGUE_TEST_INI_FUGUE_CONF = _extract_fugue_conf(conf)


def _get_all_ini_conf() -> Dict[str, Any]:
    return _FUGUE_TEST_ALL_INI_CONF


@run_once
def _load_all_backends() -> None:
    from fugue.constants import FUGUE_ENTRYPOINT

    load_entry_point(FUGUE_ENTRYPOINT)


def with_backend(ctx: Any, *other: Any, skip_missing: bool = False) -> Any:
    """The decorator to set the backend context for the test function

    :param ctx: the object to specify the backend
    :param other: more objects to specify the additional backends
    :param skip_missing: skip if an object can't be converted to a backend,
        defaults to False

    .. admonition:: Examples

        .. code-block:: python

            import fugue.test as ft
            import fugue.api as fa

            @ft.with_backend("pandas", "spark")
            def test_spark(path_fixture):
                # test load and count under spark and pandas
                df = fa.load(path_fixture)
                assert fa.count(df) == 3

    .. note::

        ``ctx`` can be a string or a tuple of string and dict. If it contains a
        dict, the dict will be merged with the default configuration for the backend.

        the configuration of each backend is extracted from the pytest ini configuration
        under ``fugue_test_conf``. The configuration must prefixed with the backend
        name.

        ``fugue_test_conf`` is defined at the parallel place where you define pytest
        ``addopts``. For example, if you define them in ``setup.cfg``, then it
        looks like:

        .. code-block:: ini

            [tool:pytest]
            addopts =
                -p pytest_cov
                --cov=fugue
            fugue_test_conf =
                fugue.test.dummy=dummy
                fugue.test:bool=true
                # ray backend settings
                ray.num_cpus:int=2
                # dask backend settings
                dask.processes:bool=true
                dask.n_workers:int=3
                dask.threads_per_worker:int=1
    """
    import pytest

    _load_all_backends()

    _ctx = _construct_parameterized_fixture([ctx] + list(other), skip_missing)
    return lambda f: pytest.mark.parametrize("backend_context", _ctx, indirect=True)(
        pytest.mark.usefixtures("backend_context")(f)
    )


def fugue_test_backend(cls: Type["FugueTestBackend"]) -> Type["FugueTestBackend"]:
    """The decorator to register a Fugue test backend. Most Fugue users don't need
    to use this decorator.

    :param cls: the FugueTestBackend class

    .. admonition:: Examples

        .. code-block:: python

            import fugue.test as ft

            @ft.fugue_test_backend
            class MyBackend(ft.FugueTestBackend):
                name = "my_backend"  # the name of the backend

                @classmethod
                @contextmanager
                def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
                    # create a session object
                    yield session

    """
    assert_or_throw(
        issubclass(cls, FugueTestBackend),
        ValueError(f"{cls} is not a FugueTestBackend"),
    )
    name = cls.name.strip().lower()
    assert_or_throw(
        name != "" and name != "fugue",
        ValueError(f"Fugue test backend name cannot be empty or fugue: {cls}"),
    )
    assert_or_throw(
        name not in _FUGUE_TEST_BACKENDS,
        ValueError(f"Duplicate Fugue test backend name: {name}"),
    )
    _FUGUE_TEST_BACKENDS[name] = cls
    return cls


class FugueTestSuite:
    """The base class for Fugue test suite. Most Fugue users don't need to use this
    class or the decorator :func:`~.fugue_test_suite`. The dccorator and the class
    are often used together.

    .. admonition:: Examples

        .. code-block:: python

            import fugue.test as ft

            @ft.fugue_test_suite("pandas")
            class MyTest(ft.FugueTestSuite):
                def test_pandas(self):
                    # test pandas
                    pass

                def test_spark(self):
                    # test spark
                    pass
    """

    backend: Any
    tmp_path: Path
    equal_type_groups: Any = None

    __test__ = False
    _test_context: Any = None

    if _HAS_PYTEST:

        @pytest.fixture(autouse=True)
        def init_builtin_per_func_context(self, tmp_path):
            self.tmp_path = tmp_path

    @property
    def context(self) -> "FugueTestContext":
        """The ``FugueTestContext`` object"""
        return self._test_context

    @property
    def engine(self) -> Any:
        """The engine object inside the ``FugueTestContext``"""
        return self.context.engine

    def get_equal_type_groups(self) -> Optional[List[List[Any]]]:
        return None  # pragma: no cover

    def df_eq(self, *args: Any, **kwargs: Any) -> bool:
        """A wrapper of :func:`~fugue.dataframe.utils.df_eq`"""
        if "equal_type_groups" not in kwargs:
            kwargs["equal_type_groups"] = self.equal_type_groups
        return _df_eq(*args, **kwargs)


def fugue_test_suite(backend: Any, mark_test: Optional[bool] = None) -> Any:
    def deco(cls: Type["FugueTestSuite"]) -> Type["FugueTestSuite"]:
        import pytest

        assert_or_throw(
            issubclass(cls, FugueTestSuite),
            ValueError(f"{cls} is not a FugueTestSuite"),
        )
        if mark_test is not None:
            cls.__test__ = mark_test
        c, extra_conf = _parse_backend(backend)
        return pytest.mark.parametrize(
            "backend_context", [pytest.param((c, extra_conf), id=c)], indirect=True
        )(pytest.mark.usefixtures("_class_backend_context")(cls))

    return deco


@dataclass
class FugueTestContext:
    """The context object for Fugue test

    :param engine: the Fugue ExecutionEngine object
    :param session: the original session wrapped by the Fugue ExecutionEngine
    :param name: the backend name
    """

    engine: Any
    session: Any
    name: str


class FugueTestBackend:
    """The base class for Fugue test backend. Most Fugue users don't need to use this
    class or the decorator :func:`~.fugue_test_backend`. The dccorator and the class
    are often used together.

    .. admonition:: Examples

        .. code-block:: python

            import fugue.test as ft

            @ft.fugue_test_backend
            class MyBackend(ft.FugueTestBackend):
                name = "my_backend"  # the name of the backend

                @classmethod
                @contextmanager
                def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
                    # create a session object
                    yield session
    """

    name = ""
    default_session_conf: Dict[str, Any] = {}
    default_fugue_conf: Dict[str, Any] = {}
    session_conf: Dict[str, Any] = {}
    fugue_conf: Dict[str, Any] = {}

    @classmethod
    def transform_session_conf(cls, conf: Dict[str, Any]) -> Dict[str, Any]:
        """Extract and transform the ``conf`` and keep the ones specific to
        this backend. The default implementation will extract the configuration
        prefixed with ``<backend name>.`` and remove the prefix.

        :param conf: the raw configuration
        :return: the transformed configuration
        """
        return extract_conf(conf, cls.name + ".", remove_prefix=True)

    @classmethod
    @contextmanager
    def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
        """Create the backend native session object

        :param session_conf: the configuration for this backend
        :yield: the session object
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    @contextmanager
    def generate_session_fixture(cls) -> Iterator[Any]:
        """Generate the session fixture for this backend from the default backend
        configuration and then the pytest ini configuration under ``fugue_test_conf``

        :yield: the session object
        """
        session_conf = _merge_dicts(
            cls.default_session_conf,
            cls.transform_session_conf(_FUGUE_TEST_ALL_INI_CONF),
        )
        with cls.session_context(session_conf) as session:
            yield session

    @classmethod
    @contextmanager
    def generate_context_fixture(
        cls, session: object, extra_fugue_conf: Dict[str, Any]
    ) -> Iterator[FugueTestContext]:
        """Generate the ``FugueTestContext`` fixture for this backend from the
        default backend configuration and then the pytest ini configuration under
        ``fugue_test_conf`` and then ``extra_fugue_conf``.

        :param session: the session object
        :param extra_fugue_conf: the extra fugue configuration

        :yield: the ``FugueTestContext`` object
        """
        import fugue.api as fa

        fugue_conf = _merge_dicts(
            cls.default_fugue_conf,
            _FUGUE_TEST_INI_FUGUE_CONF,
            _extract_fugue_conf(extra_fugue_conf),
        )
        with fa.engine_context(session, fugue_conf) as engine:
            yield FugueTestContext(engine=engine, session=session, name=cls.name)


def extract_conf(
    conf: Dict[str, Any], prefix: str, remove_prefix: bool
) -> Dict[str, Any]:
    """Extract the configuration prefixed with ``prefix`` and optionally remove the
    prefix

    :param conf: the raw configuration
    :param prefix: the prefix
    :param remove_prefix: whether to remove the prefix

    :return: the extracted configuration
    """
    res: Dict[str, Any] = {}
    for k, v in conf.items():
        if k.startswith(prefix):
            if remove_prefix:
                k = k[len(prefix) :]
            res[k] = v
    return res


@contextmanager
def _make_backend_context(obj: Any, session: Any) -> Iterator[Any]:
    _load_all_backends()
    key, extra_conf = _parse_backend(obj)
    assert_or_throw(
        key in _FUGUE_TEST_BACKENDS,
        lambda: ValueError(
            f"Undefined Fugue test backend: {key}, "
            f"available backends: {list(_FUGUE_TEST_BACKENDS.keys())}"
        ),
    )
    with _FUGUE_TEST_BACKENDS[key].generate_context_fixture(session, extra_conf) as ctx:
        yield ctx


def _extract_fugue_conf(conf: Dict[str, Any]) -> Dict[str, Any]:
    return extract_conf(conf, "fugue.", remove_prefix=False)


def _construct_parameterized_fixture(ctx: List[Any], skip_missing: bool) -> List[Any]:
    import pytest

    _ctx: List[Tuple[str, Dict[str, Any]]] = []
    for x in ctx:
        c, extra_conf = _parse_backend(x)
        if c not in _FUGUE_TEST_BACKENDS:
            if not skip_missing:
                raise ValueError(
                    f"Undefined Fugue test backend: {c}, "
                    f"available backends: {list(_FUGUE_TEST_BACKENDS.keys())}"
                )
            else:
                _ctx.append(
                    pytest.param(
                        c,
                        marks=pytest.mark.xfail(
                            reason="Undefined Fugue test backend", run=False
                        ),
                        id=c,
                    )
                )
        else:
            _ctx.append(pytest.param((c, extra_conf), id=c))
    return _ctx


def _merge_dicts(*dicts: Dict[str, Any]) -> Dict[str, Any]:
    res: Dict[str, Any] = {}
    for d in dicts:
        res.update(d)
    return res


def _parse_backend(ctx: Any) -> Tuple[str, Dict[str, Any]]:
    if isinstance(ctx, str):
        return ctx, {}
    else:
        return ctx[0], ctx[1]



from typing import Any, Optional, Union

import dask.dataframe as dd
import duckdb
import pyarrow as pa
from dask.distributed import Client
from duckdb import DuckDBPyConnection
from triad import assert_or_throw

from fugue import DataFrame, MapEngine, PartitionSpec
from fugue_dask import DaskDataFrame, DaskExecutionEngine
from fugue_dask.execution_engine import DaskMapEngine
from .dataframe import DuckDataFrame
from .execution_engine import DuckExecutionEngine, _to_duck_df


class DuckDaskExecutionEngine(DuckExecutionEngine):
    """A hybrid engine of DuckDB and Dask. Most operations will be done by
    DuckDB, but for ``map``, it will use Dask to fully utilize local CPUs.
    The engine can be used with a real Dask cluster, but practically, this
    is more useful for local process.

    :param conf: |ParamsLikeObject|, read |FugueConfig| to learn Fugue specific options
    :param connection: DuckDB connection
    """

    def __init__(
        self,
        conf: Any = None,
        connection: Optional[DuckDBPyConnection] = None,
        dask_client: Optional[Client] = None,
    ):
        super().__init__(conf, connection)
        self._dask_engine = DaskExecutionEngine(dask_client=dask_client, conf=conf)

    def create_default_map_engine(self) -> MapEngine:
        return DaskMapEngine(self._dask_engine)

    def get_current_parallelism(self) -> int:
        return self._dask_engine.get_current_parallelism()

    @property
    def dask_client(self) -> Client:
        return self._dask_engine.dask_client

    def to_df(self, df: Any, schema: Any = None) -> DuckDataFrame:
        if isinstance(df, (dd.DataFrame, DaskDataFrame)):
            ddf = self._to_dask_df(df, schema)
            if all(not pa.types.is_nested(f.type) for f in ddf.schema.fields):
                res = DuckDataFrame(self.connection.from_df(ddf.as_pandas()))
            else:
                res = DuckDataFrame(
                    duckdb.from_arrow(ddf.as_arrow(), connection=self.connection)
                )
            if ddf.has_metadata:  # pragma: no cover
                res.reset_metadata(ddf.metadata)
            return res
        return super().to_df(df, schema)

    def repartition(self, df: DataFrame, partition_spec: PartitionSpec) -> DataFrame:
        tdf = self._to_auto_df(df)
        if isinstance(tdf, DaskDataFrame):
            return self._dask_engine.repartition(tdf, partition_spec=partition_spec)
        return super().repartition(tdf, partition_spec=partition_spec)

    def broadcast(self, df: DataFrame) -> DataFrame:
        if isinstance(df, DaskDataFrame):
            return self._dask_engine.broadcast(df)
        return super().broadcast(df)

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        if isinstance(df, DaskDataFrame):
            return self._dask_engine.persist(df, lazy=lazy, **kwargs)
        return super().persist(df, lazy=lazy, **kwargs)

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        if isinstance(df, DaskDataFrame) or not partition_spec.empty:
            return self._dask_engine.save_df(
                self._to_dask_df(df),
                path=path,
                format_hint=format_hint,
                mode=mode,
                partition_spec=partition_spec,
                force_single=force_single,
                **kwargs,
            )
        return super().save_df(
            df,
            path=path,
            format_hint=format_hint,
            mode=mode,
            partition_spec=partition_spec,
            force_single=force_single,
            **kwargs,
        )

    def convert_yield_dataframe(self, df: DataFrame, as_local: bool) -> DataFrame:
        if isinstance(df, DaskDataFrame):
            return self._dask_engine.convert_yield_dataframe(df, as_local)
        return super().convert_yield_dataframe(df, as_local)

    def _to_auto_df(
        self, df: Any, schema: Any = None
    ) -> Union[DuckDataFrame, DaskDataFrame]:
        if isinstance(df, (DuckDataFrame, DaskDataFrame)):
            assert_or_throw(
                schema is None,
                ValueError("schema must be None when df is a DataFrame"),
            )
            return df
        if isinstance(df, dd.DataFrame):
            return self._dask_engine.to_df(df, schema)
        return _to_duck_df(self, df, schema)

    def _to_dask_df(self, df: Any, schema: Any = None) -> DaskDataFrame:
        if isinstance(df, DuckDataFrame):
            res = self._dask_engine.to_df(df.as_pandas(), df.schema)
            res.reset_metadata(df.metadata if df.has_metadata else None)
            return res
        return self._dask_engine.to_df(df, schema)



import logging
import os
from typing import Any, Callable, Dict, List, Optional, Type, Union

import numpy as np
import pandas as pd
from triad import Schema
from triad.collections.dict import IndexedOrderedDict
from triad.utils.assertion import assert_or_throw
from triad.utils.io import makedirs
from triad.utils.pandas_like import PandasUtils

from fugue._utils.io import load_df, save_df
from fugue._utils.misc import import_fsql_dependency
from fugue.collections.partition import (
    PartitionCursor,
    PartitionSpec,
    parse_presort_exp,
)
from fugue.collections.sql import StructuredRawSQL
from fugue.constants import KEYWORD_PARALLELISM, KEYWORD_ROWCOUNT
from fugue.dataframe import (
    AnyDataFrame,
    DataFrame,
    DataFrames,
    LocalBoundedDataFrame,
    LocalDataFrame,
    PandasDataFrame,
    fugue_annotated_param,
)
from fugue.dataframe.dataframe import as_fugue_df
from fugue.dataframe.utils import get_join_schemas

from .execution_engine import (
    ExecutionEngine,
    ExecutionEngineParam,
    MapEngine,
    SQLEngine,
)


class QPDPandasEngine(SQLEngine):
    """QPD execution implementation.

    :param execution_engine: the execution engine this sql engine will run on
    """

    @property
    def dialect(self) -> Optional[str]:
        return "spark"

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> DataFrame:
        return _to_native_execution_engine_df(df, schema)

    @property
    def is_distributed(self) -> bool:
        return False

    def select(self, dfs: DataFrames, statement: StructuredRawSQL) -> DataFrame:
        qpd_pandas = import_fsql_dependency("qpd_pandas")

        _dfs, _sql = self.encode(dfs, statement)
        _dd = {k: self.to_df(v).as_pandas() for k, v in _dfs.items()}  # type: ignore

        df = qpd_pandas.run_sql_on_pandas(_sql, _dd, ignore_case=True)
        return self.to_df(df)


class PandasMapEngine(MapEngine):
    @property
    def execution_engine_constraint(self) -> Type[ExecutionEngine]:
        return NativeExecutionEngine

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> DataFrame:
        return _to_native_execution_engine_df(df, schema)

    @property
    def is_distributed(self) -> bool:
        return False

    def map_dataframe(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        # if partition_spec.num_partitions != "0":
        #     self.log.warning(
        #         "%s doesn't respect num_partitions %s",
        #         self,
        #         partition_spec.num_partitions,
        #     )
        is_coarse = partition_spec.algo == "coarse"
        presort = partition_spec.get_sorts(df.schema, with_partition_keys=is_coarse)
        presort_keys = list(presort.keys())
        presort_asc = list(presort.values())
        output_schema = Schema(output_schema)
        cursor = partition_spec.get_cursor(df.schema, 0)
        if on_init is not None:
            on_init(0, df)
        if (
            len(partition_spec.partition_by) == 0 or partition_spec.algo == "coarse"
        ):  # no partition
            if len(partition_spec.presort) > 0:
                pdf = (
                    df.as_pandas()
                    .sort_values(presort_keys, ascending=presort_asc)
                    .reset_index(drop=True)
                )
                input_df: LocalDataFrame = PandasDataFrame(
                    pdf, df.schema, pandas_df_wrapper=True
                )
            else:
                input_df = df.as_local()
            if (
                len(partition_spec.partition_by) == 0
                and partition_spec.num_partitions != "0"
            ):
                partitions = partition_spec.get_num_partitions(
                    **{
                        KEYWORD_ROWCOUNT: lambda: df.count(),  # type: ignore
                        KEYWORD_PARALLELISM: lambda: 1,
                    }
                )
                dfs: List[DataFrame] = []
                for p, subdf in enumerate(
                    np.array_split(input_df.as_pandas(), partitions)
                ):
                    if len(subdf) > 0:
                        tdf = PandasDataFrame(subdf, df.schema, pandas_df_wrapper=True)
                        cursor.set(lambda: tdf.peek_array(), p, 0)
                        dfs.append(map_func(cursor, tdf).as_pandas())
                output_df: LocalDataFrame = PandasDataFrame(
                    pd.concat(dfs, ignore_index=True),
                    schema=output_schema,
                    pandas_df_wrapper=True,
                )
            else:
                cursor.set(lambda: input_df.peek_array(), 0, 0)
                output_df = map_func(cursor, input_df)
            if (
                isinstance(output_df, PandasDataFrame)
                and output_df.schema != output_schema
            ):  # pragma: no cover
                output_df = PandasDataFrame(output_df.native, output_schema)
            assert_or_throw(
                output_df.schema == output_schema,
                lambda: f"map output {output_df.schema} "
                f"mismatches given {output_schema}",
            )
            return self.to_df(output_df)  # type: ignore

        def _map(pdf: pd.DataFrame) -> pd.DataFrame:
            if len(partition_spec.presort) > 0:
                pdf = pdf.sort_values(presort_keys, ascending=presort_asc).reset_index(
                    drop=True
                )
            input_df = PandasDataFrame(pdf, df.schema, pandas_df_wrapper=True)
            cursor.set(lambda: input_df.peek_array(), cursor.partition_no + 1, 0)
            output_df = map_func(cursor, input_df)
            return output_df.as_pandas()

        result = self.execution_engine.pl_utils.safe_groupby_apply(  # type: ignore
            df.as_pandas(), partition_spec.partition_by, _map
        )
        return PandasDataFrame(result, output_schema)


class NativeExecutionEngine(ExecutionEngine):
    """The execution engine based on native python and pandas. This execution engine
    is mainly for prototyping and unit tests.

    Please read |ExecutionEngineTutorial| to understand this important Fugue concept

    :param conf: |ParamsLikeObject|, read |FugueConfig| to learn Fugue specific options
    """

    def __init__(self, conf: Any = None):
        super().__init__(conf)
        self._log = logging.getLogger()

    def __repr__(self) -> str:
        return "NativeExecutionEngine"

    @property
    def log(self) -> logging.Logger:
        return self._log

    @property
    def is_distributed(self) -> bool:
        return False

    def create_default_sql_engine(self) -> SQLEngine:
        return QPDPandasEngine(self)

    def create_default_map_engine(self) -> MapEngine:
        return PandasMapEngine(self)

    def get_current_parallelism(self) -> int:
        return 1

    @property
    def pl_utils(self) -> PandasUtils:
        """Pandas-like dataframe utils"""
        return PandasUtils()

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> LocalBoundedDataFrame:
        return _to_native_execution_engine_df(df, schema)  # type: ignore

    def repartition(
        self, df: DataFrame, partition_spec: PartitionSpec
    ) -> DataFrame:  # pragma: no cover
        # self.log.warning("%s doesn't respect repartition", self)
        return df

    def broadcast(self, df: DataFrame) -> DataFrame:
        return self.to_df(df)

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        return self.to_df(df)

    def join(
        self,
        df1: DataFrame,
        df2: DataFrame,
        how: str,
        on: Optional[List[str]] = None,
    ) -> DataFrame:
        key_schema, output_schema = get_join_schemas(df1, df2, how=how, on=on)
        d = self.pl_utils.join(
            df1.as_pandas(), df2.as_pandas(), join_type=how, on=key_schema.names
        )
        return PandasDataFrame(d.reset_index(drop=True), output_schema)

    def union(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.union(df1.as_pandas(), df2.as_pandas(), unique=distinct)
        return PandasDataFrame(d.reset_index(drop=True), df1.schema)

    def subtract(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            distinct, NotImplementedError("EXCEPT ALL for NativeExecutionEngine")
        )
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.except_df(df1.as_pandas(), df2.as_pandas(), unique=distinct)
        return PandasDataFrame(d.reset_index(drop=True), df1.schema)

    def intersect(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            distinct, NotImplementedError("INTERSECT ALL for NativeExecutionEngine")
        )
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.intersect(df1.as_pandas(), df2.as_pandas(), unique=distinct)
        return PandasDataFrame(d.reset_index(drop=True), df1.schema)

    def distinct(
        self,
        df: DataFrame,
    ) -> DataFrame:
        d = self.pl_utils.drop_duplicates(df.as_pandas())
        return PandasDataFrame(d.reset_index(drop=True), df.schema)

    def dropna(
        self,
        df: DataFrame,
        how: str = "any",
        thresh: Optional[int] = None,
        subset: List[str] = None,
    ) -> DataFrame:
        kwargs: Dict[str, Any] = dict(axis=0, subset=subset, inplace=False)
        if thresh is None:
            kwargs["how"] = how
        else:
            kwargs["thresh"] = thresh
        d = df.as_pandas().dropna(**kwargs)
        return PandasDataFrame(d.reset_index(drop=True), df.schema)

    def fillna(
        self,
        df: DataFrame,
        value: Any,
        subset: List[str] = None,
    ) -> DataFrame:
        assert_or_throw(
            (not isinstance(value, list)) and (value is not None),
            ValueError("fillna value can not None or a list"),
        )
        if isinstance(value, dict):
            assert_or_throw(
                (None not in value.values()) and (any(value.values())),
                ValueError(
                    "fillna dict can not contain None and needs at least one value"
                ),
            )
            mapping = value
        else:
            # If subset is none, apply to all columns
            subset = subset or df.columns
            mapping = {col: value for col in subset}
        d = df.as_pandas().fillna(mapping, inplace=False)
        return PandasDataFrame(d.reset_index(drop=True), df.schema)

    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:
        assert_or_throw(
            (n is None and frac is not None) or (n is not None and frac is None),
            ValueError("one and only one of n and frac should be set"),
        )
        d = df.as_pandas().sample(n=n, frac=frac, replace=replace, random_state=seed)
        return PandasDataFrame(d.reset_index(drop=True), df.schema)

    def take(
        self,
        df: DataFrame,
        n: int,
        presort: str,
        na_position: str = "last",
        partition_spec: Optional[PartitionSpec] = None,
    ) -> DataFrame:
        partition_spec = partition_spec or PartitionSpec()
        assert_or_throw(
            isinstance(n, int),
            ValueError("n needs to be an integer"),
        )
        d = df.as_pandas()

        # Use presort over partition_spec.presort if possible
        if presort:
            presort = parse_presort_exp(presort)
        _presort: IndexedOrderedDict = presort or partition_spec.presort

        if len(_presort.keys()) > 0:
            d = d.sort_values(
                list(_presort.keys()),
                ascending=list(_presort.values()),
                na_position=na_position,
            )

        if len(partition_spec.partition_by) == 0:
            d = d.head(n)
        else:
            d = d.groupby(by=partition_spec.partition_by, dropna=False).head(n)

        return PandasDataFrame(
            d.reset_index(drop=True), df.schema, pandas_df_wrapper=True
        )

    def load_df(
        self,
        path: Union[str, List[str]],
        format_hint: Any = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> LocalBoundedDataFrame:
        return self.to_df(
            load_df(path, format_hint=format_hint, columns=columns, **kwargs)
        )

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        if not force_single and not partition_spec.empty:
            kwargs["partition_cols"] = partition_spec.partition_by
        makedirs(os.path.dirname(path), exist_ok=True)
        df = self.to_df(df)
        save_df(df, path, format_hint=format_hint, mode=mode, **kwargs)


@fugue_annotated_param(NativeExecutionEngine)
class _NativeExecutionEngineParam(ExecutionEngineParam):
    pass


def _to_native_execution_engine_df(df: AnyDataFrame, schema: Any = None) -> DataFrame:
    fdf = as_fugue_df(df) if schema is None else as_fugue_df(df, schema=schema)
    return fdf.as_local_bounded()



import sys
from abc import ABC, abstractmethod
from types import TracebackType
from typing import Any, Callable, List, Optional, no_type_check

from adagio.instances import TaskContext
from adagio.specs import InputSpec, OutputSpec, TaskSpec
from triad import ParamDict
from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_or_throw
from triad.utils.hash import to_uuid

from fugue._utils.exception import (
    _MODIFIED_EXCEPTION_VAR_NAME,
    frames_to_traceback,
    modify_traceback,
)
from fugue.collections.partition import PartitionSpec
from fugue.collections.yielded import PhysicalYielded
from fugue.dataframe import DataFrame, DataFrames
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.exceptions import FugueWorkflowError
from fugue.execution import ExecutionEngine
from fugue.extensions.creator.convert import _to_creator
from fugue.extensions.outputter.convert import _to_outputter
from fugue.extensions.processor.convert import _to_processor
from fugue.rpc.base import RPCServer
from fugue.workflow._checkpoint import Checkpoint, StrongCheckpoint
from fugue.workflow._workflow_context import FugueWorkflowContext


class FugueTask(TaskSpec, ABC):
    def __init__(
        self,
        input_n: int = 0,
        output_n: int = 0,
        configs: Any = None,
        params: Any = None,
        deterministic: bool = True,
        lazy: bool = False,
        input_names: Optional[List[str]] = None,
    ):
        assert_or_throw(
            output_n <= 1,  # TODO: for now we don't support multi output
            NotImplementedError("Fugue doesn't support multiple output tasks"),
        )
        if input_names is None:
            inputs = [
                InputSpec("_" + str(i), DataFrame, nullable=False)
                for i in range(input_n)
            ]
        else:
            inputs = [
                InputSpec(input_names[i], DataFrame, nullable=False)
                for i in range(input_n)
            ]
        outputs = [
            OutputSpec("_" + str(i), DataFrame, nullable=False) for i in range(output_n)
        ]
        self._input_has_key = input_names is not None
        super().__init__(
            configs=configs,
            inputs=inputs,
            outputs=outputs,
            func=self.execute,
            metadata=params,
            deterministic=deterministic,
            lazy=lazy,
        )
        self._checkpoint = Checkpoint()
        self._broadcast = False
        self._dependency_uuid: Any = None
        self._yield_dataframe_handler: Any = None
        self._yield_local_dataframe = False
        self._traceback: Optional[TracebackType] = None

    def reset_traceback(
        self, limit: int, should_prune: Optional[Callable[[str], bool]] = None
    ) -> None:
        cf = sys._getframe(1)
        self._traceback = frames_to_traceback(
            cf, limit=limit, should_prune=should_prune
        )

    def __uuid__(self) -> str:
        # _checkpoint is not part of determinism
        # _yield_name is not part of determinism
        return to_uuid(
            self.configs,
            self.inputs,
            self.outputs,
            # get_full_type_path(self.func),
            self.metadata,
            self.deterministic,
            self.lazy,
            self._get_dependency_uuid(),
            self._broadcast,
        )

    @abstractmethod
    def execute(self, ctx: TaskContext) -> None:  # pragma: no cover
        raise NotImplementedError

    @property
    def single_output_expression(self) -> str:
        assert_or_throw(
            len(self.outputs) == 1,
            lambda: FugueWorkflowError(f"{self.name} does not have single output"),
        )
        return self.name + "." + self.outputs.get_key_by_index(0)

    def copy(self) -> "FugueTask":
        raise InvalidOperationError("can't copy")

    def __copy__(self) -> "FugueTask":
        raise InvalidOperationError("can't copy")

    def __deepcopy__(self, memo: Any) -> "FugueTask":
        raise InvalidOperationError("can't copy")

    def set_checkpoint(self, checkpoint: Checkpoint) -> "FugueTask":
        self._checkpoint = checkpoint
        return self

    @property
    def has_checkpoint(self) -> bool:
        return not self._checkpoint.is_null

    @property
    def yielded(self) -> PhysicalYielded:
        if isinstance(self._checkpoint, StrongCheckpoint):
            return self._checkpoint.yielded
        raise ValueError("can't populate value from non-physical yield")

    def broadcast(self) -> "FugueTask":
        self._broadcast = True
        return self

    def set_yield_dataframe_handler(self, handler: Callable, as_local: bool) -> None:
        self._yield_dataframe_handler = handler
        self._yield_local_dataframe = as_local

    def set_result(self, ctx: TaskContext, df: DataFrame) -> DataFrame:
        df = self._handle_checkpoint(df, ctx)
        df = self._handle_broadcast(df, ctx)
        if self._yield_dataframe_handler is not None:
            out_df = self._get_execution_engine(ctx).convert_yield_dataframe(
                df, as_local=self._yield_local_dataframe
            )
            self._yield_dataframe_handler(out_df)
        self._get_workflow_context(ctx).set_result(id(self), df)
        return df

    def _get_workflow_context(self, ctx: TaskContext) -> FugueWorkflowContext:
        wfctx = ctx.workflow_context
        assert isinstance(wfctx, FugueWorkflowContext)
        return wfctx

    def _get_execution_engine(self, ctx: TaskContext) -> ExecutionEngine:
        return self._get_workflow_context(ctx).execution_engine

    def _get_rpc_server(self, ctx: TaskContext) -> RPCServer:
        return self._get_workflow_context(ctx).rpc_server

    def _handle_checkpoint(self, df: DataFrame, ctx: TaskContext) -> DataFrame:
        wfctx = self._get_workflow_context(ctx)
        return self._execute_with_modified_traceback(
            lambda: self._checkpoint.run(df, wfctx.checkpoint_path)
        )

    def _handle_broadcast(self, df: DataFrame, ctx: TaskContext) -> DataFrame:
        if not self._broadcast:
            return df
        return self._execute_with_modified_traceback(
            lambda: self._get_execution_engine(ctx).broadcast(df)
        )

    def _get_dependency_uuid(self) -> Any:
        # TODO: this should be a part of adagio!!
        if self._dependency_uuid is not None:
            return self._dependency_uuid
        values: List[Any] = []
        for k, v in self.node_spec.dependency.items():
            t = v.split(".", 1)
            assert_or_throw(len(t) == 2)
            values.append(k)
            values.append(t[1])
            task = self.parent_workflow.tasks[t[0]]
            values.append(task.__uuid__())
        self._dependency_uuid = to_uuid(values)
        return self._dependency_uuid

    def _execute_with_modified_traceback(self, func: Callable) -> Any:
        try:
            return func()
        except Exception as ex:
            if self._traceback is None:  # pragma: no cover
                raise

            if _MODIFIED_EXCEPTION_VAR_NAME in self._traceback.tb_frame.f_locals:
                raise ex from self._traceback.tb_frame.f_locals[
                    _MODIFIED_EXCEPTION_VAR_NAME
                ]

            # add caller traceback
            ctb = modify_traceback(
                sys.exc_info()[2].tb_next, None, self._traceback  # type: ignore
            )
            if ctb is None:  # pragma: no cover
                raise
            raise ex.with_traceback(ctb)


class Create(FugueTask):
    @no_type_check
    def __init__(
        self,
        creator: Any,
        schema: Any = None,
        params: Any = None,
        deterministic: bool = True,
        lazy: bool = True,
    ):
        self._creator = _to_creator(creator, schema)
        self._creator._params = ParamDict(params, deep=False)
        super().__init__(
            params=params, input_n=0, output_n=1, deterministic=deterministic, lazy=lazy
        )

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(super().__uuid__(), self._creator, self._creator._params)

    @no_type_check
    def execute(self, ctx: TaskContext) -> None:
        e = self._get_execution_engine(ctx)
        self._creator._execution_engine = e
        df = self._execute_with_modified_traceback(lambda: self._creator.create())
        df = self.set_result(ctx, df)
        ctx.outputs["_0"] = df


class Process(FugueTask):
    @no_type_check
    def __init__(
        self,
        input_n: int,
        processor: Any,
        schema: Any,
        params: Any,
        pre_partition: Any = None,
        deterministic: bool = True,
        lazy: bool = False,
        input_names: Optional[List[str]] = None,
    ):
        self._processor = _to_processor(processor, schema)
        self._processor._params = ParamDict(params)
        self._processor._partition_spec = PartitionSpec(pre_partition)
        self._processor.validate_on_compile()
        super().__init__(
            params=params,
            input_n=input_n,
            output_n=1,
            deterministic=deterministic,
            lazy=lazy,
            input_names=input_names,
        )

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(
            super().__uuid__(),
            self._processor,
            self._processor._params,
            self._processor._partition_spec,
        )

    @no_type_check
    def execute(self, ctx: TaskContext) -> None:
        e = self._get_execution_engine(ctx)
        self._processor._execution_engine = e
        self._processor._rpc_server = self._get_rpc_server(ctx)
        if self._input_has_key:
            inputs = DataFrames(ctx.inputs)
        else:
            inputs = DataFrames(ctx.inputs.values())

        def exe() -> Any:
            self._processor.validate_on_runtime(inputs)
            return self._processor.process(inputs)

        df = self._execute_with_modified_traceback(exe)
        df = self.set_result(ctx, df)
        ctx.outputs["_0"] = df


class Output(FugueTask):
    @no_type_check
    def __init__(
        self,
        input_n: int,
        outputter: Any,
        params: Any,
        pre_partition: Any = None,
        deterministic: bool = True,
        lazy: bool = False,
        input_names: Optional[List[str]] = None,
    ):
        assert_or_throw(input_n > 0, FugueWorkflowError("must have at least one input"))
        self._outputter = _to_outputter(outputter)
        self._outputter._params = ParamDict(params)
        self._outputter._partition_spec = PartitionSpec(pre_partition)
        self._outputter.validate_on_compile()
        super().__init__(
            params=params,
            input_n=input_n,
            output_n=1,
            deterministic=deterministic,
            lazy=lazy,
            input_names=input_names,
        )

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(
            super().__uuid__(),
            self._outputter,
            self._outputter._params,
            self._outputter._partition_spec,
        )

    @no_type_check
    def execute(self, ctx: TaskContext) -> None:
        self._outputter._execution_engine = self._get_execution_engine(ctx)
        self._outputter._rpc_server = self._get_rpc_server(ctx)
        if self._input_has_key:
            inputs = DataFrames(ctx.inputs)
        else:
            inputs = DataFrames(ctx.inputs.values())

        def exe():
            self._outputter.validate_on_runtime(inputs)
            self._outputter.process(inputs)

        self._execute_with_modified_traceback(exe)
        # TODO: output dummy to force cache to work, should we fix adagio?
        ctx.outputs["_0"] = ArrayDataFrame([], "_0:int")



from abc import abstractmethod
from typing import Any, Dict, Iterable, List, Optional

import pandas as pd
import pyarrow as pa
from triad import Schema, assert_or_throw

from fugue import DataFrame, IterableDataFrame, LocalBoundedDataFrame
from fugue.dataframe.dataframe import _input_schema
from fugue.exceptions import FugueDataFrameOperationError, FugueDatasetEmptyError
from fugue.plugins import drop_columns, get_column_names, is_df, rename

from ._compat import IbisSchema, IbisTable
from ._utils import pa_to_ibis_type, to_schema


class IbisDataFrame(DataFrame):
    """DataFrame that wraps Ibis ``Table``.

    :param rel: ``DuckDBPyRelation`` object
    """

    def __init__(self, table: IbisTable, schema: Any = None):
        if schema is not None:
            _schema = _input_schema(schema).assert_not_empty()
        else:
            _schema = self._to_schema(table.schema())
        self._table = self._alter_table_columns(table, _schema)
        super().__init__(schema=_schema)

    @abstractmethod
    def to_sql(self) -> str:  # pragma: no cover
        """Compile IbisTable to SQL"""
        raise NotImplementedError

    @property
    def native(self) -> IbisTable:
        """Ibis Table object"""
        return self._table

    def native_as_df(self) -> IbisTable:
        return self._table

    def _to_schema(self, schema: IbisSchema) -> Schema:
        return to_schema(schema)

    def _to_local_df(
        self, table: IbisTable, schema: Any = None
    ) -> LocalBoundedDataFrame:
        raise NotImplementedError  # pragma: no cover

    def _to_iterable_df(
        self, table: IbisTable, sdhema: Any = None
    ) -> IterableDataFrame:
        raise NotImplementedError  # pragma: no cover

    def _to_new_df(self, table: IbisTable, schema: Any = None) -> DataFrame:
        raise NotImplementedError  # pragma: no cover

    def _compute_scalar(self, table: IbisTable) -> Any:
        return table.execute()

    @property
    def is_local(self) -> bool:
        return False

    @property
    def is_bounded(self) -> bool:
        return True

    @property
    def empty(self) -> bool:
        return self._to_local_df(self._table.limit(1)).count() == 0

    @property
    def num_partitions(self) -> int:
        return 1  # pragma: no cover

    @property
    def columns(self) -> List[str]:
        return _get_ibis_columns(self._table)

    def peek_array(self) -> List[Any]:
        res = self._to_local_df(self._table.head(1)).as_array()
        if len(res) == 0:
            raise FugueDatasetEmptyError()
        return res[0]

    def count(self) -> int:
        return self._compute_scalar(self._table.count())

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        schema = self.schema.exclude(cols)
        return self._to_new_df(self._table[schema.names], schema=schema)

    def _select_cols(self, keys: List[Any]) -> DataFrame:
        schema = self.schema.extract(keys)
        return self._to_new_df(self._table[schema.names], schema=schema)

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            schema = self.schema.rename(columns)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        df = _rename(self._table, self.columns, schema.names)
        return self if df is self._table else self._to_new_df(df, schema=schema)

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self.schema.alter(columns)
        if new_schema == self.schema:
            return self
        return self._to_new_df(
            self._alter_table_columns(self._table, new_schema),
            schema=new_schema,
        )

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        return self.as_local().as_arrow(type_safe=type_safe)

    def as_pandas(self) -> pd.DataFrame:
        return self.as_local().as_pandas()

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        res = self._to_local_df(self._table, schema=self.schema)
        if res is not self and self.has_metadata:
            res.reset_metadata(self.metadata)
        return res

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        if columns is not None:
            return self[columns].as_array(type_safe=type_safe)
        return self.as_local().as_array(type_safe=type_safe)

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        if columns is not None:
            yield from self[columns].as_array_iterable(type_safe=type_safe)
        else:
            yield from self._to_iterable_df(self._table).as_array_iterable(
                type_safe=type_safe
            )

    def as_dicts(self, columns: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        if columns is not None:
            return self[columns].as_dicts()
        return self.as_local().as_dicts()

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None
    ) -> Iterable[Dict[str, Any]]:
        if columns is not None:
            yield from self[columns].as_dict_iterable()
        else:
            yield from self._to_iterable_df(self._table).as_dict_iterable()

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        if columns is not None:
            return self[columns].head(n)
        return self._to_local_df(self._table.head(n)).as_local_bounded()

    def _alter_table_columns(self, table: IbisTable, new_schema: Schema) -> IbisTable:
        fields: Dict[str, Any] = {}
        schema = table.schema()
        for _name, _type, f2 in zip(schema.names, schema.types, new_schema.fields):
            _new_name, _new_type = f2.name, pa_to_ibis_type(f2.type)
            assert_or_throw(
                _name == _new_name,
                lambda: ValueError(f"schema name mismatch: {_name} vs {_new_name}"),
            )
            if _type == _new_type:
                continue
            else:
                fields[_name] = table[_name].cast(_new_type)
        if len(fields) == 0:
            return table
        return table.mutate(**fields)


@is_df.candidate(lambda df: isinstance(df, IbisTable))
def _ibis_is_df(df: IbisTable) -> bool:
    return True


@get_column_names.candidate(lambda df: isinstance(df, IbisTable))
def _get_ibis_columns(df: IbisTable) -> List[Any]:
    return df.columns


@drop_columns.candidate(lambda df, *args, **kwargs: isinstance(df, IbisTable))
def _drop_ibis_columns(df: IbisTable, columns: List[str]) -> IbisTable:
    cols = [x for x in df.columns if x not in columns]
    if len(cols) == 0:
        raise FugueDataFrameOperationError("cannot drop all columns")
    if len(cols) + len(columns) != len(df.columns):
        _assert_no_missing(df, columns)
    return df[cols]


@rename.candidate(lambda df, *args, **kwargs: isinstance(df, IbisTable))
def _rename_ibis_table(df: IbisTable, columns: Dict[str, Any]) -> IbisTable:
    _assert_no_missing(df, columns.keys())
    old_names = df.columns
    new_names = [columns.get(name, name) for name in old_names]
    return _rename(df, old_names, new_names)


def _rename(df: IbisTable, old_names: List[str], new_names: List[str]) -> IbisTable:
    cols: List[Any] = []
    has_change = False
    for a, b in zip(old_names, new_names):
        if a == b:
            cols.append(df[a])
        else:
            cols.append(df[a].name(b))
            has_change = True
    return df.projection(cols) if has_change else df


def _assert_no_missing(df: IbisTable, columns: Iterable[Any]) -> None:
    missing = set(columns) - set(df.columns)
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")



from typing import Any, Dict, Iterable, List, Optional, Tuple

import pandas as pd
import pyarrow as pa
import ray
import ray.data as rd
from triad import assert_or_throw
from triad.collections.schema import Schema
from triad.utils.pyarrow import cast_pa_table

from fugue.dataframe import ArrowDataFrame, DataFrame, LocalBoundedDataFrame
from fugue.dataframe.dataframe import _input_schema
from fugue.dataframe.utils import pa_table_as_array, pa_table_as_dicts
from fugue.exceptions import FugueDataFrameOperationError, FugueDatasetEmptyError
from fugue.plugins import (
    as_array,
    as_array_iterable,
    as_arrow,
    as_dict_iterable,
    as_dicts,
    as_local_bounded,
    as_pandas,
    get_column_names,
    get_num_partitions,
    is_df,
    rename,
)

from ._constants import _ZERO_COPY
from ._utils.dataframe import build_empty, get_dataset_format, materialize, to_schema


class RayDataFrame(DataFrame):
    """DataFrame that wraps Ray DataSet. Please also read
    |DataFrameTutorial| to understand this Fugue concept

    :param df: :class:`ray:ray.data.Dataset`, :class:`pa:pyarrow.Table`,
      :class:`pd:pandas.DataFrame`,
      Fugue :class:`~fugue.dataframe.dataframe.DataFrame`,
      or list or iterable of arrays
    :param schema: |SchemaLikeObject|, defaults to None. If the schema
      is different from the ``df`` schema, then type casts will happen.
    :param internal_schema: for internal schema, it means the schema
      is guaranteed by the provider to be consistent with the schema of
      ``df``, so no type cast will happen. Defaults to False. This is
      for internal use only.
    """

    def __init__(  # noqa: C901
        self,
        df: Any = None,
        schema: Any = None,
        internal_schema: bool = False,
    ):
        metadata: Any = None
        if internal_schema:
            schema = _input_schema(schema).assert_not_empty()
        if df is None:
            schema = _input_schema(schema).assert_not_empty()
            super().__init__(schema)
            self._native = build_empty(schema)
            return
        if isinstance(df, rd.Dataset):
            fmt, df = get_dataset_format(df)
            if fmt is None:  # empty:
                schema = _input_schema(schema).assert_not_empty()
                super().__init__(schema)
                self._native = build_empty(schema)
                return
            elif fmt == "pandas":
                rdf = rd.from_arrow_refs(df.to_arrow_refs())
            elif fmt == "arrow":
                rdf = df
            else:  # pragma: no cover
                raise NotImplementedError(
                    f"Ray Dataset in {fmt} format is not supported"
                )
        elif isinstance(df, pa.Table):
            rdf = rd.from_arrow(df)
            if schema is None:
                schema = df.schema
        elif isinstance(df, RayDataFrame):
            rdf = df._native
            if schema is None:
                schema = df.schema
            metadata = None if not df.has_metadata else df.metadata
        elif isinstance(df, (pd.DataFrame, pd.Series)):
            if isinstance(df, pd.Series):
                df = df.to_frame()
            adf = ArrowDataFrame(df)
            rdf = rd.from_arrow(adf.native)
            if schema is None:
                schema = adf.schema
        elif isinstance(df, Iterable):
            schema = _input_schema(schema).assert_not_empty()
            t = ArrowDataFrame(df, schema)
            rdf = rd.from_arrow(t.as_arrow())
        elif isinstance(df, DataFrame):
            rdf = rd.from_arrow(df.as_arrow(type_safe=True))
            if schema is None:
                schema = df.schema
            metadata = None if not df.has_metadata else df.metadata
        else:
            raise ValueError(f"{df} is incompatible with RayDataFrame")
        rdf, schema = self._apply_schema(rdf, schema, internal_schema)
        super().__init__(schema)
        self._native = rdf
        if metadata is not None:  # pragma: no cover
            self.reset_metadata(metadata)

    @property
    def native(self) -> rd.Dataset:
        """The wrapped ray Dataset"""
        return self._native

    def native_as_df(self) -> rd.Dataset:
        return self._native

    @property
    def is_local(self) -> bool:
        return False

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        adf = self.as_arrow()
        if adf.shape[0] == 0:
            res = ArrowDataFrame([], self.schema)
        else:
            res = ArrowDataFrame(adf)
        if self.has_metadata:
            res.reset_metadata(self.metadata)
        return res

    @property
    def is_bounded(self) -> bool:
        return True

    @property
    def empty(self) -> bool:
        return len(self.native.take(1)) == 0

    @property
    def num_partitions(self) -> int:
        return _rd_num_partitions(self.native)

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        cols = (self.schema - cols).names
        return self._select_cols(cols)

    def _select_cols(self, cols: List[Any]) -> DataFrame:
        if cols == self.columns:
            return self
        return RayDataFrame(
            self.native.select_columns(cols),
            self.schema.extract(cols),
            internal_schema=True,
        )

    def peek_array(self) -> List[Any]:
        data = self.native.limit(1).to_pandas().values.tolist()
        if len(data) == 0:
            raise FugueDatasetEmptyError
        return data[0]

    def persist(self, **kwargs: Any) -> "RayDataFrame":
        # TODO: it mutates the dataframe, is this a good bahavior
        self._native = materialize(self._native)
        return self

    def count(self) -> int:
        return self.native.count()

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        return _rd_as_arrow(self.native)

    def as_pandas(self) -> pd.DataFrame:
        return _rd_as_pandas(self.native)

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            new_schema = self.schema.rename(columns)
            new_cols = new_schema.names
        except Exception as e:
            raise FugueDataFrameOperationError from e
        rdf = self.native.map_batches(
            lambda b: b.rename_columns(new_cols),
            batch_format="pyarrow",
            **_ZERO_COPY,
            **self._remote_args(),
        )
        return RayDataFrame(rdf, schema=new_schema, internal_schema=True)

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self.schema.alter(columns)

        def _alter(
            table: pa.Table,
        ) -> pa.Table:  # pragma: no cover (pytest can't capture)
            return cast_pa_table(table, new_schema.pa_schema)

        if self.schema == new_schema:
            return self
        rdf = self.native.map_batches(
            _alter, batch_format="pyarrow", **_ZERO_COPY, **self._remote_args()
        )
        return RayDataFrame(rdf, schema=new_schema, internal_schema=True)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return _rd_as_array(self.native, columns, type_safe)

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        yield from _rd_as_array_iterable(self.native, columns, type_safe)

    def as_dicts(self, columns: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        return _rd_as_dicts(self.native, columns)

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None
    ) -> Iterable[Dict[str, Any]]:
        yield from _rd_as_dict_iterable(self.native, columns)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        if columns is not None:
            return self[columns].head(n)
        pdf = RayDataFrame(self.native.limit(n), schema=self.schema)
        return pdf.as_local()  # type: ignore

    def _apply_schema(
        self, rdf: rd.Dataset, schema: Optional[Schema], internal_schema: bool
    ) -> Tuple[rd.Dataset, Schema]:
        if internal_schema:
            return rdf, schema
        fmt, rdf = get_dataset_format(rdf)
        if fmt is None:  # empty
            schema = _input_schema(schema).assert_not_empty()
            return build_empty(schema), schema
        if schema is None or schema == to_schema(rdf.schema(fetch_if_missing=True)):
            return rdf, to_schema(rdf.schema(fetch_if_missing=True))

        def _alter(table: pa.Table) -> pa.Table:  # pragma: no cover
            return ArrowDataFrame(table).alter_columns(schema).native  # type: ignore

        return (
            rdf.map_batches(
                _alter, batch_format="pyarrow", **_ZERO_COPY, **self._remote_args()
            ),
            schema,
        )

    def _remote_args(self) -> Dict[str, Any]:
        return {"num_cpus": 1}


@is_df.candidate(lambda df: isinstance(df, rd.Dataset))
def _rd_is_df(df: rd.Dataset) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, rd.Dataset))
def _rd_num_partitions(df: rd.Dataset) -> int:
    return df.num_blocks()


@as_local_bounded.candidate(lambda df: isinstance(df, rd.Dataset))
def _rd_as_local(df: rd.Dataset) -> pa.Table:
    return _rd_as_arrow(df)


@get_column_names.candidate(lambda df: isinstance(df, rd.Dataset))
def _get_ray_dataframe_columns(df: rd.Dataset) -> List[Any]:
    if hasattr(df, "columns"):  # higher version of ray
        return df.columns(fetch_if_missing=True)
    else:  # pragma: no cover
        fmt, _ = get_dataset_format(df)
        if fmt == "pandas":
            return list(df.schema(True).names)
        elif fmt == "arrow":
            return df.schema(fetch_if_missing=True).names
        raise NotImplementedError(f"{fmt} is not supported")  # pragma: no cover


@rename.candidate(lambda df, *args, **kwargs: isinstance(df, rd.Dataset))
def _rename_ray_dataframe(df: rd.Dataset, columns: Dict[str, Any]) -> rd.Dataset:
    if len(columns) == 0:
        return df
    cols = _get_ray_dataframe_columns(df)
    missing = set(columns.keys()) - set(cols)
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")
    new_cols = [columns.get(name, name) for name in cols]
    return df.map_batches(
        lambda b: b.rename_columns(new_cols), batch_format="pyarrow", **_ZERO_COPY
    )


@as_pandas.candidate(lambda df: isinstance(df, rd.Dataset))
def _rd_as_pandas(df: rd.Dataset) -> pd.DataFrame:
    return _rd_as_arrow(df).to_pandas()


@as_arrow.candidate(lambda df: isinstance(df, rd.Dataset))
def _rd_as_arrow(df: rd.Dataset) -> pa.Table:
    return pa.concat_tables(_get_arrow_tables(df))


@as_array.candidate(lambda df, *args, **kwargs: isinstance(df, rd.Dataset))
def _rd_as_array(
    df: rd.Dataset, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Any]:
    assert_or_throw(columns is None or len(columns) > 0, ValueError("empty columns"))
    _df = df if columns is None or len(columns) == 0 else df.select_columns(columns)
    adf = _rd_as_arrow(_df)
    return pa_table_as_array(adf)


@as_array_iterable.candidate(lambda df, *args, **kwargs: isinstance(df, rd.Dataset))
def _rd_as_array_iterable(
    df: rd.Dataset, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Any]:
    yield from _rd_as_array(df, columns, type_safe)


@as_dicts.candidate(lambda df, *args, **kwargs: isinstance(df, rd.Dataset))
def _rd_as_dicts(
    df: rd.Dataset, columns: Optional[List[str]] = None, type_safe: bool = False
) -> List[Dict[str, Any]]:
    assert_or_throw(columns is None or len(columns) > 0, ValueError("empty columns"))
    _df = df if columns is None or len(columns) == 0 else df.select_columns(columns)
    adf = _rd_as_arrow(_df)
    return pa_table_as_dicts(adf)


@as_dict_iterable.candidate(lambda df, *args, **kwargs: isinstance(df, rd.Dataset))
def _rd_as_dict_iterable(
    df: rd.Dataset, columns: Optional[List[str]] = None, type_safe: bool = False
) -> Iterable[Dict[str, Any]]:
    yield from _rd_as_dicts(df, columns, type_safe)


def _get_arrow_tables(df: rd.Dataset) -> Iterable[pa.Table]:
    last_empty: Any = None
    empty = True
    for block in df.to_arrow_refs():
        tb = ray.get(block)
        if tb.shape[0] > 0:
            yield tb
            empty = False
        else:
            last_empty = tb
    if empty:
        yield last_empty



from typing import Any, Optional

import dask.dataframe as dd

try:
    from dask.dataframe.dask_expr.io.parquet import ReadParquet

    HAS_DASK_EXPR = True  # newer dask
except ImportError:  # pragma: no cover
    HAS_DASK_EXPR = False  # older dask

if not HAS_DASK_EXPR:  # pragma: no cover
    try:
        from dask_sql import Context as ContextWrapper  # pylint: disable-all
    except ImportError:  # pragma: no cover
        raise ImportError(
            "dask-sql is not installed. Please install it with `pip install dask-sql`"
        )
else:
    from triad.utils.assertion import assert_or_throw

    try:
        from dask_sql import Context
        from dask_sql.datacontainer import Statistics
        from dask_sql.input_utils import InputUtil
    except ImportError:  # pragma: no cover
        raise ImportError(
            "dask-sql is not installed. Please install it with `pip install dask-sql`"
        )

    class ContextWrapper(Context):  # type: ignore
        def create_table(
            self,
            table_name: str,
            input_table: dd.DataFrame,
            format: Optional[str] = None,  # noqa
            persist: bool = False,
            schema_name: Optional[str] = None,
            statistics: Optional[Statistics] = None,
            gpu: bool = False,
            **kwargs: Any,
        ) -> None:  # pragma: no cover
            assert_or_throw(
                isinstance(input_table, dd.DataFrame),
                lambda: ValueError(
                    f"input_table must be a dask dataframe, but got {type(input_table)}"
                ),
            )
            assert_or_throw(
                dd._dask_expr_enabled(), lambda: ValueError("Dask expr must be enabled")
            )
            schema_name = schema_name or self.schema_name

            dc = InputUtil.to_dc(
                input_table,
                table_name=table_name,
                format=format,
                persist=persist,
                gpu=gpu,
                **kwargs,
            )

            dask_filepath = None
            operations = input_table.find_operations(ReadParquet)
            for op in operations:
                dask_filepath = op._args[0]

            dc.filepath = dask_filepath
            self.schema[schema_name].filepaths[table_name.lower()] = dask_filepath

            if not statistics:
                statistics = Statistics(float("nan"))
            dc.statistics = statistics

            self.schema[schema_name].tables[table_name.lower()] = dc
            self.schema[schema_name].statistics[table_name.lower()] = statistics



import pytest

_DEFAULT_SCOPE = "module"


@pytest.fixture(scope=_DEFAULT_SCOPE)
def pandas_session():
    yield "pandas"


@pytest.fixture(scope=_DEFAULT_SCOPE)
def native_session():
    yield "native"


@pytest.fixture(scope=_DEFAULT_SCOPE)
def dask_session():
    from fugue_dask.tester import DaskTestBackend

    with DaskTestBackend.generate_session_fixture() as session:
        yield session


@pytest.fixture(scope=_DEFAULT_SCOPE)
def duckdb_session():
    from fugue_duckdb.tester import DuckDBTestBackend

    with DuckDBTestBackend.generate_session_fixture() as session:
        yield session


@pytest.fixture(scope=_DEFAULT_SCOPE)
def duckdask_session():
    from fugue_duckdb.tester import DuckDaskTestBackend

    with DuckDaskTestBackend.generate_session_fixture() as session:
        yield session


@pytest.fixture(scope=_DEFAULT_SCOPE)
def ray_session():
    from fugue_ray.tester import RayTestBackend

    with RayTestBackend.generate_session_fixture() as session:
        yield session


@pytest.fixture(scope=_DEFAULT_SCOPE)
def spark_session():
    from fugue_spark.tester import SparkTestBackend

    with SparkTestBackend.generate_session_fixture() as session:
        yield session


@pytest.fixture(scope=_DEFAULT_SCOPE)
def sparkconnect_session():
    from fugue_spark.tester import SparkConnectTestBackend

    with SparkConnectTestBackend.generate_session_fixture() as session:
        yield session



# flake8: noqa
from .pandas_tester import NativeTestBackend, PandasTestBackend
from .plugins import (
    FugueTestBackend,
    FugueTestContext,
    FugueTestSuite,
    extract_conf,
    fugue_test_backend,
    fugue_test_suite,
    with_backend,
)



from typing import Any, Dict
from uuid import uuid4

from adagio.instances import (
    NoOpCache,
    ParallelExecutionEngine,
    WorkflowContext,
    WorkflowHooks,
)
from adagio.specs import WorkflowSpec
from fugue.constants import FUGUE_CONF_WORKFLOW_CONCURRENCY
from fugue.dataframe import DataFrame
from fugue.execution.execution_engine import ExecutionEngine
from fugue.rpc.base import make_rpc_server, RPCServer
from fugue.workflow._checkpoint import CheckpointPath
from triad import SerializableRLock, ParamDict


class FugueWorkflowContext(WorkflowContext):
    def __init__(
        self,
        engine: ExecutionEngine,
        compile_conf: Any = None,
        cache: Any = NoOpCache,
        workflow_engine: Any = None,
        hooks: Any = WorkflowHooks,
    ):
        conf = ParamDict(compile_conf)
        self._fugue_engine = engine
        self._lock = SerializableRLock()
        self._results: Dict[Any, DataFrame] = {}
        self._execution_id = ""
        self._checkpoint_path = CheckpointPath(self.execution_engine)
        self._rpc_server = make_rpc_server(engine.conf)
        if workflow_engine is None:
            workflow_engine = ParallelExecutionEngine(
                conf.get_or_throw(FUGUE_CONF_WORKFLOW_CONCURRENCY, int),
                self,
            )
        super().__init__(
            cache=cache,
            engine=workflow_engine,
            hooks=hooks,
            logger=self.execution_engine.log,
            config=conf,
        )

    def run(self, spec: WorkflowSpec, conf: Dict[str, Any]) -> None:
        try:
            self._execution_id = str(uuid4())
            self._checkpoint_path = CheckpointPath(self.execution_engine)
            self._checkpoint_path.init_temp_path(self._execution_id)
            self._rpc_server.start()
            super().run(spec, conf)
        finally:
            self._checkpoint_path.remove_temp_path()
            self._rpc_server.stop()
            self._execution_id = ""

    @property
    def checkpoint_path(self) -> CheckpointPath:
        return self._checkpoint_path

    @property
    def execution_engine(self) -> ExecutionEngine:
        return self._fugue_engine

    @property
    def rpc_server(self) -> RPCServer:
        return self._rpc_server

    def set_result(self, key: Any, df: DataFrame) -> None:
        with self._lock:
            self._results[key] = df

    def get_result(self, key: Any) -> DataFrame:
        with self._lock:
            return self._results[key]



import os
import pathlib
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import pyarrow as pa
import ray.data as rd
from packaging import version
from pyarrow import csv as pacsv
from pyarrow import json as pajson

from triad.collections import Schema
from triad.collections.dict import ParamDict
from triad.utils.assertion import assert_or_throw
from triad.utils.io import exists, makedirs, rm, isfile

from fugue import ExecutionEngine
from fugue._utils.io import FileParser, save_df
from fugue.collections.partition import PartitionSpec
from fugue.dataframe import DataFrame
from fugue_ray.dataframe import RayDataFrame

from .._constants import RAY_VERSION

try:
    from ray.data.datasource import FileExtensionFilter

    class _FileFiler(FileExtensionFilter):  # pragma: no cover
        def __init__(
            self, file_extensions: Union[str, List[str]], exclude: Iterable[str]
        ):
            super().__init__(file_extensions, allow_if_no_extension=True)
            self._exclude = set(exclude)

        def _is_valid(self, path: str) -> bool:
            return pathlib.Path(
                path
            ).name not in self._exclude and self._file_has_extension(path)

        def __call__(self, paths: List[str]) -> List[str]:
            return [path for path in paths if self._is_valid(path)]

except ImportError:  # pragma: no cover
    pass  # ray >=2.10


class RayIO(object):
    def __init__(self, engine: ExecutionEngine):
        self._engine = engine
        self._logger = engine.log
        self._loads: Dict[str, Callable[..., DataFrame]] = {
            "csv": self._load_csv,
            "parquet": self._load_parquet,
            "json": self._load_json,
        }
        self._saves: Dict[str, Callable[..., None]] = {
            "csv": self._save_csv,
            "parquet": self._save_parquet,
            "json": self._save_json,
        }

    def load_df(
        self,
        uri: Union[str, List[str]],
        format_hint: Optional[str] = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> DataFrame:
        if isinstance(uri, str):
            fp = [FileParser(uri, format_hint)]
        else:
            fp = [FileParser(u, format_hint) for u in uri]
        fmts = list(set(f.file_format for f in fp))  # noqa: C401
        assert_or_throw(
            len(fmts) == 1, NotImplementedError("can't support multiple formats")
        )
        fmt = fmts[0]
        files = [f.as_dir_path() if f.is_dir else f.path for f in fp]
        return self._loads[fmt](files, columns, **kwargs)

    def save_df(
        self,
        df: RayDataFrame,
        uri: str,
        format_hint: Optional[str] = None,
        partition_spec: Optional[PartitionSpec] = None,
        mode: str = "overwrite",
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        if exists(uri):
            assert_or_throw(mode == "overwrite", FileExistsError(uri))
            try:
                rm(uri, recursive=True)
            except Exception:  # pragma: no cover
                pass
        p = FileParser(uri, format_hint)
        if not force_single:
            df = self._prepartition(df, partition_spec=partition_spec)

            self._saves[p.file_format](df=df, uri=p.path, **kwargs)
        else:
            ldf = df.as_local()
            makedirs(os.path.dirname(uri), exist_ok=True)
            save_df(ldf, uri, format_hint=format_hint, mode=mode, **kwargs)

    def _save_parquet(
        self,
        df: RayDataFrame,
        uri: str,
        **kwargs: Any,
    ) -> None:
        df.native.write_parquet(uri, ray_remote_args=self._remote_args(), **kwargs)

    def _save_csv(
        self,
        df: RayDataFrame,
        uri: str,
        **kwargs: Any,
    ) -> None:
        kw = dict(kwargs)
        if "header" in kw:
            kw["include_header"] = kw.pop("header")

        def _fn() -> Dict[str, Any]:  # pragma: no cover
            return dict(write_options=pacsv.WriteOptions(**kw))

        df.native.write_csv(
            uri, ray_remote_args=self._remote_args(), arrow_csv_args_fn=_fn
        )

    def _save_json(
        self,
        df: RayDataFrame,
        uri: str,
        **kwargs: Any,
    ) -> None:
        df.native.write_json(uri, ray_remote_args=self._remote_args(), **kwargs)

    def _prepartition(
        self, rdf: RayDataFrame, partition_spec: PartitionSpec
    ) -> RayDataFrame:
        # will not use bucket by because here we don't save as table
        by = partition_spec.partition_by
        if len(by) > 0:  # pragma: no cover
            self._logger.warning(
                f"prepartitioning by keys {by} is not supported by ray, will ignore"
            )
        return self._engine.repartition(
            rdf, partition_spec=partition_spec  # type: ignore
        )

    def _load_parquet(
        self, p: List[str], columns: Any = None, **kwargs: Any
    ) -> DataFrame:
        # in 2.52.0 the default changes to ["parquet"]
        if "file_extensions" not in kwargs:
            kwargs = kwargs.copy()
            kwargs["file_extensions"] = None
        sdf = rd.read_parquet(p, ray_remote_args=self._remote_args(), **kwargs)
        if columns is None:
            return RayDataFrame(sdf)
        if isinstance(columns, list):  # column names
            return RayDataFrame(sdf)[columns]
        schema = Schema(columns)
        return RayDataFrame(sdf)[schema.names].alter_columns(schema)

    def _load_csv(  # noqa: C901
        self, p: List[str], columns: Any = None, **kwargs: Any
    ) -> DataFrame:
        kw = ParamDict(kwargs)
        infer_schema = kw.get("infer_schema", False)
        read_options: Dict[str, Any] = {"use_threads": False}
        parse_options: Dict[str, Any] = {}
        convert_options: Dict[str, Any] = {}
        if infer_schema and columns is not None and not isinstance(columns, list):
            raise ValueError("can't set columns as a schema when infer schema is true")

        if RAY_VERSION >= version.parse("2.10"):
            if len(p) == 1 and isfile(p[0]):  # TODO: very hacky
                params: Dict[str, Any] = {}
            else:
                params = {"file_extensions": ["csv"]}
        else:  # pragma: no cover
            params = {
                "partition_filter": _FileFiler(
                    file_extensions=["csv"], exclude=["_SUCCESS"]
                ),
            }

        def _read_csv(to_str: bool) -> RayDataFrame:
            res = rd.read_csv(
                p,
                ray_remote_args=self._remote_args(),
                read_options=pacsv.ReadOptions(**read_options),
                parse_options=pacsv.ParseOptions(**parse_options),
                convert_options=pacsv.ConvertOptions(**convert_options),
                **params,
            )
            if to_str:
                _schema = res.schema(fetch_if_missing=True)
                str_schema = Schema([(x, pa.string()) for x in _schema.names]).pa_schema
                return RayDataFrame(res, schema=str_schema)
            else:
                return RayDataFrame(res)

        header = str(kw.get_or_none("header", bool)).lower()
        if header == "true":
            read_options["autogenerate_column_names"] = False
            df = _read_csv(not infer_schema)
            if columns is None:
                return df
            if isinstance(columns, list):
                return df[columns]
            schema = Schema(columns)
            return df[schema.names].alter_columns(schema)
        elif header in ["false", "none"]:
            read_options["autogenerate_column_names"] = False
            if columns is None:
                raise ValueError("columns must be set if without header")
            if isinstance(columns, list):  # column names
                read_options["column_names"] = columns
                return _read_csv(not infer_schema)
            else:
                schema = Schema(columns)
                read_options["column_names"] = schema.names
                convert_options["column_types"] = schema.pa_schema
                return _read_csv(False)
        else:  # pragma: no cover
            raise NotImplementedError(f"{header} is not supported")

    def _load_json(self, p: List[str], columns: Any = None, **kwargs: Any) -> DataFrame:
        read_options: Dict[str, Any] = {"use_threads": False}
        parse_options: Dict[str, Any] = {}

        def _read_json() -> RayDataFrame:  # pragma: no cover
            if RAY_VERSION >= version.parse("2.10"):
                if len(p) == 1 and isfile(p[0]):  # TODO: very hacky
                    params: Dict[str, Any] = {"file_extensions": None}
                else:
                    params = {"file_extensions": ["json"]}
            elif RAY_VERSION >= version.parse("2.9"):  # pragma: no cover
                params = {
                    "file_extensions": None,
                    "partition_filter": _FileFiler(
                        file_extensions=["json"], exclude=["_SUCCESS"]
                    ),
                }
            else:  # pragma: no cover
                params = {
                    "partition_filter": _FileFiler(
                        file_extensions=["json"], exclude=["_SUCCESS"]
                    ),
                }
            return RayDataFrame(
                rd.read_json(
                    p,
                    ray_remote_args=self._remote_args(),
                    read_options=pajson.ReadOptions(**read_options),
                    parse_options=pajson.ParseOptions(**parse_options),
                    **params,
                )
            )

        if columns is None or isinstance(columns, list):
            rdf = _read_json()
            if isinstance(columns, list):  # column names
                return rdf[columns]
            return rdf
        else:
            schema = Schema(columns)
            return _read_json()[schema.names].alter_columns(schema)

    def _remote_args(self) -> Dict[str, Any]:
        return {"num_cpus": 1}



from typing import Any

import ray.data as rd
from triad import run_at_def

from fugue import DataFrame, register_execution_engine
from fugue.dev import (
    DataFrameParam,
    ExecutionEngineParam,
    fugue_annotated_param,
    is_pandas_or,
)
from fugue.plugins import as_fugue_dataset, infer_execution_engine

from .dataframe import RayDataFrame
from .execution_engine import RayExecutionEngine
from .tester import RayTestBackend  # noqa: F401  # pylint: disable-all


@infer_execution_engine.candidate(
    lambda objs: is_pandas_or(objs, (rd.Dataset, RayDataFrame))
)
def _infer_ray_client(objs: Any) -> Any:
    return "ray"


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, rd.Dataset))
def _ray_as_fugue_df(df: rd.Dataset, **kwargs: Any) -> RayDataFrame:
    return RayDataFrame(df, **kwargs)


def _register_engines() -> None:
    register_execution_engine(
        "ray", lambda conf, **kwargs: RayExecutionEngine(conf=conf), on_dup="ignore"
    )


@fugue_annotated_param(RayExecutionEngine)
class _RayExecutionEngineParam(ExecutionEngineParam):
    pass


@fugue_annotated_param(rd.Dataset)
class _RayDatasetParam(DataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        assert isinstance(ctx, RayExecutionEngine)
        return ctx._to_ray_df(df).native

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert isinstance(output, rd.Dataset)
        assert isinstance(ctx, RayExecutionEngine)
        return RayDataFrame(output, schema=schema)

    def count(self, df: DataFrame) -> int:  # pragma: no cover
        raise NotImplementedError("not allowed")


@run_at_def
def _register() -> None:
    """Register Ray Execution Engine"""
    _register_engines()



from contextlib import contextmanager
from typing import Any, Dict, Iterator

import duckdb

import fugue.test as ft

try:
    import dask.distributed as dd
    import dask

    _HAS_DASK = True
except ImportError:  # pragma: no cover
    _HAS_DASK = False


@ft.fugue_test_backend
class DuckDBTestBackend(ft.FugueTestBackend):
    name = "duckdb"

    @classmethod
    @contextmanager
    def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
        with duckdb.connect(config=session_conf) as conn:
            yield conn


if _HAS_DASK:

    @ft.fugue_test_backend
    class DuckDaskTestBackend(ft.FugueTestBackend):
        name = "duckdask"

        @classmethod
        def transform_session_conf(cls, conf: Dict[str, Any]) -> Dict[str, Any]:
            res = ft.extract_conf(conf, "duck.", remove_prefix=False)
            res.update(ft.extract_conf(conf, "dask.", remove_prefix=False))
            return res

        @classmethod
        @contextmanager
        def session_context(cls, session_conf: Dict[str, Any]) -> Iterator[Any]:
            duck_conf = ft.extract_conf(session_conf, "duck.", remove_prefix=True)
            dask_conf = ft.extract_conf(session_conf, "dask.", remove_prefix=True)
            with dd.Client(**dask_conf) as client:
                dask.config.set({"dataframe.shuffle.method": "tasks"})
                dask.config.set({"dataframe.convert-string": False})
                with duckdb.connect(config=duck_conf) as conn:
                    yield [conn, client]
